import hashlib
import secrets
from uuid import uuid4

import rulez
from morpfw.crud import Collection, Model, Schema

from ..app import App
from .schema import APIKeySchema


class APIKeyModel(Model):
    schema = APIKeySchema
    update_view_enabled = False

    @property
    def client_id(self):
        return self["api_identity"]

    def generate_secret(self):
        api_secret = secrets.token_urlsafe(32)
        pwhash = hashlib.sha256(api_secret.encode("utf-8")).hexdigest()
        self.update({"api_secret": pwhash})
        return api_secret

    def validate(self, api_secret):
        pwhash = hashlib.sha256(api_secret.encode("utf-8")).hexdigest()
        if pwhash == self["api_secret"]:
            return True
        return False

    def user(self):
        users = self.request.get_collection("morpfw.pas.user")
        return users.get_by_userid(self["userid"])


class APIKeyCollection(Collection):
    schema = APIKeySchema
    create_view_enabled = False

    def search(self, query=None, *args, **kwargs):
        if kwargs.get("secure", True):
            if query:
                rulez.and_(rulez.field["userid"] == self.request.identity.userid, query)
            else:
                query = rulez.field["userid"] == self.request.identity.userid
        return super(APIKeyCollection, self).search(query, *args, **kwargs)

    def get_by_identity(self, identity):
        res = self.search(rulez.field("api_identity") == identity)
        if res:
            return res[0]
        return None



import pytz

import morepath

from .authn.pas.utils import has_role


class Identity(morepath.Identity):
    def __init__(self, request, userid, **kw):
        self.request = request
        super().__init__(userid, **kw)

    def timezone(self):
        return pytz.UTC

    def is_administrator(self):
        return has_role(self.request, "administrator", userid=self.userid)



import rulez
import re

valid_identifier_pattern = re.compile(r'^[a-z][a-z0-9_]*$')
valid_namespaced_identifier_pattern = re.compile(r'^[a-z][a-z0-9_\.]*$')

def unique(request, schema, field, value, mode=None):
    if mode=='update':
        return 

    typeinfo = request.app.config.type_registry.get_typeinfo_by_schema(
            schema=schema, request=request)
    collection = typeinfo['collection_factory'](request)
    if collection.search(rulez.field[field] == value):
        return "Already exists"

def valid_identifier(request, schema, field, value, mode=None):
    if not valid_identifier_pattern.match(value):
        return (
            "Only lower cased alphanumeric and _ are accepted and shall"
            " not start with number nor _")


def valid_namespaced_identifier(request, schema, field, value, mode=None):
    if not valid_namespaced_identifier_pattern.match(value):
        return (
            "Only lower cased alphanumeric, . and _ are accepted, and shall"
            " not start with number, . nor _")





import sys

import click
from alembic.config import CommandLine as AlembicCLI
from alembic.config import Config as AlembicCfg
from alembic.config import main as alembic_main

from . import db, elasticstore, register_admin, scheduler, shell, start, worker
from .cli import cli
from .generate_config import genconfig as genconfig_main


@cli.command(help="generate config file")
@click.pass_context
def generate_config(ctx, options):
    pass


def main():
    if "generate-config" in sys.argv:
        argv = sys.argv[sys.argv.index("generate-config") + 1 :]
        sys.exit(genconfig_main(argv))
    else:
        cli()


if __name__ == "__main__":
    main()




import typing
import morpfw
from .app import App
from .model import PageCollection


class PagesSearchProvider(morpfw.SearchProvider):

    def search(self, query=None, offset=0, limit=None, order_by=None):
        """search for resources and return list of resource model objects"""
        result = []
        # do something here
        return result


@App.searchprovider(model=PageCollection)
def get_searchprovider(context):
    return PagesSearchProvider(context)



from ..errors import UnprocessableError
from ..model import Collection
from ...interfaces import ISearchProvider
from rulez import parse_dsl, OperatorNotAllowedError


class SearchProvider(ISearchProvider):

    def __init__(self, context: Collection):
        self.context = context
        self.storage = context.storage
        self.request = context.request

    def parse_query(self, qs):
        if not qs.strip():
            return None
        try:
            return parse_dsl(qs)
        except ValueError:
            raise UnprocessableError("Invalid search query '%s'" % qs)
        except OperatorNotAllowedError:
            raise UnprocessableError("Invalid search query '%s'" % qs)
        return None



import typing

import dectate
import morepath
import reg
from morepath.directive import SettingAction

from . import signals
from .blobstorage.base import BlobStorage
from .typeregistry import TypeRegistry


class StorageAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return str((self.model))

    def perform(self, obj, app_class):
        app_class._get_storage.register(
            reg.methodify(obj),
            model=self.model,
            request=morepath.Request,
            blobstorage=BlobStorage,
        )


class BlobStorageAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return str((self.model))

    def perform(self, obj, app_class):
        app_class._get_blobstorage.register(
            reg.methodify(obj), model=self.model, request=morepath.Request
        )


class DataProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema, obj, storage):
        self.schema = schema
        self.obj = obj
        self.storage = storage

    def identifier(self, app_class):
        return (self.schema, self.obj, self.storage)

    def perform(self, obj, app_class):
        app_class.get_dataprovider.register(
            reg.methodify(obj), schema=self.schema, obj=self.obj, storage=self.storage
        )


class JSONProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, obj):
        self.obj = obj

    def identifier(self, app_class):
        return (self.obj,)

    def perform(self, obj, app_class):
        app_class.get_jsonprovider.register(reg.methodify(obj), obj=self.obj)


class IdentifierFieldAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema):
        self.schema = schema

    def identifier(self, app_class):
        return (self.schema,)

    def perform(self, obj, app_class):
        app_class.get_identifierfield.register(reg.methodify(obj), schema=self.schema)


class UUIDFieldAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema):
        self.schema = schema

    def identifier(self, app_class):
        return (self.schema,)

    def perform(self, obj, app_class):
        app_class.get_uuidfield.register(reg.methodify(obj), schema=self.schema)


class DefaultIdentifierAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema):
        self.schema = schema

    def identifier(self, app_class):
        return (self.schema,)

    def perform(self, obj, app_class):
        app_class.get_default_identifier.register(
            reg.methodify(obj), schema=self.schema
        )


class FormValidatorAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema):
        self.schema = schema

    def identifier(self, app_class):
        return (self.schema,)

    def perform(self, obj, app_class):
        app_class.get_formvalidators.register(reg.methodify(obj), schema=self.schema)


class RulesProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return (self.model,)

    def perform(self, obj, app_class):
        app_class.get_rulesprovider.register(reg.methodify(obj), model=self.model)


class StateMachineAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return (self.model,)

    def perform(self, obj, app_class):
        app_class.get_statemachine.register(reg.methodify(obj), model=self.model)

        def factory(model):
            return obj

        app_class.get_statemachine_factory.register(
            reg.methodify(factory), model=self.model
        )


class SearchProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return (self.model,)

    def perform(self, obj, app_class):
        app_class.get_searchprovider.register(reg.methodify(obj), model=self.model)


class AggregateProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return (self.model,)

    def perform(self, obj, app_class):
        app_class.get_aggregateprovider.register(reg.methodify(obj), model=self.model)


class XattrProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return (self.model,)

    def perform(self, obj, app_class):
        app_class.get_xattrprovider.register(reg.methodify(obj), model=self.model)


class TypeInfoFactoryAction(dectate.Action):

    config = {"type_registry": TypeRegistry}

    app_class_arg = True

    depends = [SettingAction]

    def __init__(self, name, schema):
        self.name = name
        self.schema = schema

    def identifier(self, app_class, type_registry: TypeRegistry):
        return self.name

    def perform(self, obj, app_class, type_registry: TypeRegistry):
        def factory(name):
            return obj

        app_class.get_typeinfo_factory.register(reg.methodify(factory), name=self.name)
        type_registry.register_type(name=self.name, schema=self.schema)


class StorageFactoryAction(dectate.Action):

    app_class_arg = True

    def __init__(self, name):
        self.name = name

    def identifier(self, app_class):
        return self.name

    def perform(self, obj, app_class):
        def factory(name):
            return obj

        app_class._get_storage_factory.register(reg.methodify(factory), name=self.name)


class BlobStorageFactoryAction(dectate.Action):

    app_class_arg = True

    def __init__(self, name):
        self.name = name

    def identifier(self, app_class):
        return self.name

    def perform(self, obj, app_class):
        def factory(name):
            return obj

        app_class._get_blobstorage_factory.register(
            reg.methodify(factory), name=self.name
        )


class MetalinkAction(dectate.Action):

    app_class_arg = True

    def __init__(self, name, model):
        self.name = name
        self.model = model

    def identifier(self, app_class):
        return self.name

    def perform(self, obj, app_class):
        def name_factory(name, request):
            return obj(request)

        app_class._get_metalinkprovider_by_name.register(
            reg.methodify(name_factory), name=self.name
        )

        def model_factory(model, request):
            return obj(request)

        app_class._get_metalinkprovider_by_model.register(
            reg.methodify(model_factory), model=self.model
        )


class AuthzRuleAction(dectate.Action):

    app_class_arg = True

    def __init__(self, name):
        self.name = name

    def identifier(self, app_class):
        return self.name

    def perform(self, obj, app_class):
        def factory(name):
            return obj

        app_class.get_authz_rule.register(reg.methodify(factory), name=self.name)




from morpfw.crud import signals
from .app import App
from .model import PageModel

MYSIGNAL = 'my_signal'


@App.view(model=PageModel, name='dispatch')
def dispatch(request, context):
    request.app.dispatcher(MYSIGNAL).dispatch(request, context)


@App.subscribe(model=PageModel, signal=MYSIGNAL)
def handle_signal(request, context, signal):
    print("hello world!")



import typing
import morpfw
from .app import App
from .model import PageModel


class PageRulesProvider(morpfw.RulesProvider):

    def calculate_value_offset(self):
        return self.context['value'] + 1


@App.rulesprovider(model=PageModel)
def get_rulesprovider(context):
    return PageRulesProvider(context)


@App.json(model=PageModel, name='get-value-offset')
def get_value_offset(context, request):
    return context.rulesprovider().calculate_value_offset()



import os
import pprint
import typing
from dataclasses import dataclass, field

import morpfw.crud.signals as signals
from elasticsearch import Elasticsearch
from more.basicauth import BasicAuthIdentityPolicy
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request
from morpfw.crud.blobstorage.fsblobstorage import FSBlobStorage
from morpfw.crud.model import Collection, Model
from morpfw.crud.schema import Schema
from morpfw.crud.storage.elasticsearchstorage import ElasticSearchStorage
from morpfw.request import ESCapableRequest
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register as register_session

from ..common import get_client
from .crud_common import FSBLOB_DIR
from .crud_common import App as BaseApp
from .crud_common import (
    BlobObjectCollection,
    BlobObjectModel,
    NamedObjectCollection,
    NamedObjectModel,
)
from .crud_common import ObjectSchema as BaseObjectSchema
from .crud_common import (
    ObjectXattrProvider,
    ObjectXattrSchema,
    PageCollection,
    PageModel,
    PageSchema,
    run_jslcrud_test,
)

Session = sessionmaker()
register_session(Session)


class App(BaseApp, TransactionApp):

    request_class = ESCapableRequest


class PageStorage(ElasticSearchStorage):
    index_name = "test-page"
    refresh = True
    model = PageModel


@App.path(model=PageCollection, path="pages")
def collection_factory(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path="pages/{identifier}")
def model_factory(request, identifier):
    col = collection_factory(request)
    return col.get(identifier)


@App.typeinfo(name="tests.page", schema=PageSchema)
def get_page_typeinfo(request):
    return {
        "title": "Page",
        "description": "Page type",
        "schema": PageSchema,
        "collection": PageCollection,
        "collection_factory": collection_factory,
        "model": PageModel,
        "model_factory": model_factory,
    }


@dataclass
class ObjectSchema(BaseObjectSchema):

    id: typing.Optional[str] = None


@App.identifierfield(schema=ObjectSchema)
def object_identifierfield(schema):
    return "id"


@App.default_identifier(schema=ObjectSchema)
def object_default_identifier(schema, obj, request):
    return None


class ObjectModel(Model):
    schema = ObjectSchema


@App.xattrprovider(model=ObjectModel)
def get_xattr_provider(context):
    return ObjectXattrProvider(context)


class ObjectCollection(Collection):
    schema = ObjectSchema


@App.subscribe(signal=signals.OBJECT_CREATED, model=ObjectModel)
def object_created(app, request, obj, signal):
    obj.data["created_flag"] = True


@App.subscribe(signal=signals.OBJECT_UPDATED, model=ObjectModel)
def object_updated(app, request, obj, signal):
    obj.data["updated_flag"] = True


class ObjectStorage(ElasticSearchStorage):
    index_name = "test-object"
    refresh = True
    auto_id = True
    model = ObjectModel


@App.json(model=ObjectCollection, name="get_uuid")
def get_object_by_uuid(context, request):
    uuid = request.GET.get("uuid")
    return context.get_by_uuid(uuid).json()


@App.path(model=ObjectCollection, path="objects")
def object_collection_factory(request):
    storage = ObjectStorage(request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path="objects/{identifier}")
def object_model_factory(request, identifier):
    col = object_collection_factory(request)
    return col.get(identifier)


class NamedObjectStorage(ElasticSearchStorage):
    index_name = "test-namedobject"
    refresh = True
    model = NamedObjectModel


@App.path(model=NamedObjectCollection, path="named_objects")
def namedobject_collection_factory(request):
    storage = NamedObjectStorage(request)
    return NamedObjectCollection(request, storage)


@App.path(model=NamedObjectModel, path="named_objects/{identifier}")
def namedobject_model_factory(request, identifier):
    col = namedobject_collection_factory(request)
    return col.get(identifier)


class BlobObjectStorage(ElasticSearchStorage):
    index_name = "test-blobobject"
    refresh = True
    model = BlobObjectModel


@App.storage(model=BlobObjectModel)
def get_blobobject_storage(model, request, blobstorage):
    return BlobObjectStorage(request, blobstorage=blobstorage)


@App.blobstorage(model=BlobObjectModel)
def get_blobobject_blobstorage(model, request):
    return FSBlobStorage(request, FSBLOB_DIR)


@App.path(model=BlobObjectCollection, path="blob_objects")
def blobobject_collection_factory(request):
    storage = request.app.get_storage(BlobObjectModel, request)
    return BlobObjectCollection(request, storage)


@App.path(model=BlobObjectModel, path="blob_objects/{identifier}")
def blobobject_model_factory(request, identifier):
    col = blobobject_collection_factory(request)
    return col.get(identifier)


def test_elasticsearchstorage(es_client):
    config = os.path.join(
        os.path.dirname(__file__), "test_elasticsearchstorage-settings.yml"
    )
    client = get_client(config)
    col = client.mfw_request.get_collection("tests.page")
    col.storage.create_index(col)
    run_jslcrud_test(client)



import reg
from reg.dispatch import Dispatch, validate_signature
from reg.context import DispatchMethod
from reg.predicate import PredicateRegistry
from functools import partial


def _dispatch_subscribe(self, func=None, **key_dict):
    if func is None:
        return partial(self.subscribe, **key_dict)
    validate_signature(func, self.wrapped_func)
    predicate_key = self.registry.key_dict_to_predicate_key(key_dict)
    self.registry.subscribe(predicate_key, func)
    return func


def _dispatch_publish(self, *args, **kwargs):
    subscribers = self.by_args(*args, **kwargs).all_matches
    return list([sub(*args, **kwargs) for sub in subscribers])


def _dispatchmethod_publish(self, app, *args, **kwargs):
    subscribers = self.by_args(*args, **kwargs).all_matches
    return list([sub(app, *args, **kwargs) for sub in subscribers])


def _registry_subscribe(self, key, value):
    for index, key_item in zip(self.indexes, key):
        index.setdefault(key_item, set()).add(value)


if not getattr(PredicateRegistry, '__pubsub_patched', False):
    PredicateRegistry.subscribe = _registry_subscribe
    PredicateRegistry.__pubsub_patched = True


if not getattr(Dispatch, '__pubsub_patched', False):
    Dispatch.subscribe = _dispatch_subscribe
    Dispatch.publish = _dispatch_publish
    Dispatch.__pubsub_patched = True

if not getattr(DispatchMethod, '__pubsub_dispatchmethod_patched', False):
    DispatchMethod.publish = _dispatchmethod_publish
    DispatchMethod.__pubsub_dispatchmethod_patched = True



import rulez


class Reference(object):
    def __init__(
        self,
        name: str,
        resource_type: str,
        *,
        attribute: str = "uuid",
        title=None,
        metadata=None,
    ):
        self.name = name
        self.resource_type = resource_type
        self.attribute = attribute
        self.title = title
        self.metadata = metadata or {}

    def collection(self, request):
        return request.get_collection(self.resource_type)

    def get_title(self, request):
        typeinfo = request.get_typeinfo(self.resource_type)
        return self.title or typeinfo["title"]


class ReferenceResolver(object):
    def __init__(self, request, context, reference: Reference):
        self.context = context
        self.request = request
        self.ref = reference
        self.title = reference.get_title(request)

    def resolve(self):
        value = self.context[self.ref.name]
        collection = self.ref.collection(self.request)
        items = collection.search(rulez.field(self.ref.attribute) == value)
        if items:
            return items[0]
        return None


class BackReference(object):
    def __init__(
        self,
        name: str,
        resource_type: str,
        reference_name: str,
        *,
        title=None,
        single=False,
        metadata=None,
    ):
        self.name = name
        self.resource_type = resource_type
        self.reference_name = reference_name
        self.title = title
        self.single_reference = single
        self.metadata = metadata or {}

    def collection(self, request):
        return request.get_collection(self.resource_type)

    def get_title(self, request):
        typeinfo = request.get_typeinfo(self.resource_type)
        return self.title or typeinfo["title"]

    def get_reference(self, request):
        typeinfo = request.app.config.type_registry.get_typeinfo(
            name=self.resource_type, request=request
        )
        refschema = typeinfo["schema"]
        for ref in refschema.__references__:
            if ref.name == self.reference_name:
                return ref


class BackReferenceResolver(object):
    def __init__(self, request, context, backreference: BackReference):
        self.context = context
        self.request = request
        self.bref = backreference
        self.title = backreference.get_title(request)

    def resolve(self):
        request = self.request
        context = self.context

        reference = self.bref.get_reference(request)

        if reference is None:
            raise ValueError("Invalid reference name. %s" % self.bref.reference_name)

        value = context[reference.attribute]
        collection = self.bref.collection(request)
        items = collection.search(rulez.field(reference.name) == value)
        return items




from ...app import BaseApp as App
from ...crud import permission as crudperms
from ...crud.model import Collection, Model
from ...permission import All


@App.authz_rule(name="morpfw.submit-edit")
def group_policy(groupname, identity, model, permission):
    request = model.request
    users = request.get_collection("morpfw.pas.user")
    user = users.get_by_userid(identity.userid)
    if user["is_administrator"]:
        return True

    if groupname not in [g["groupname"] for g in user.groups()]:
        return False

    if isinstance(model, Collection):
        if issubclass(permission, crudperms.View):
            return True
        if issubclass(permission, crudperms.Search):
            return True
        if issubclass(permission, crudperms.Create):
            return True
    elif isinstance(model, Model):
        if model["creator"] != user.uuid:
            return False

        if issubclass(permission, crudperms.All):
            return True

        if issubclass(permission, crudperms.StateUpdate):
            return True
    return False




import datetime
import json
import os
import typing
from dataclasses import dataclass, field
from io import BytesIO
from pathlib import Path
from urllib.parse import urlparse
from uuid import uuid4

import deform
import morepath
import morpfw
import morpfw.sql
import sqlalchemy as sa

from ..app import App
from .base import Blob, BlobStorage


@dataclass
class TypeBlobStoreSchema(morpfw.Schema):

    file: typing.Optional[deform.FileData] = None


class TypeBlobStoreMixin(object):

    filename = sa.Column(sa.String(1024))
    mimetype = sa.Column(sa.String(256))
    size = sa.Column(sa.Integer())
    encoding = sa.Column(sa.String(128))
    fp = sa.Column(sa.LargeBinary())


class TypeBlobStoreCollection(morpfw.Collection):
    schema = TypeBlobStoreSchema

    def _create(self, data):
        data = data["file"]
        data["fp"] = data["fp"].read()
        return super()._create(data)


class TypeBlobStoreModel(morpfw.Model):
    schema = TypeBlobStoreSchema

    def __setitem__(self, key, value):
        if key == "file":
            self.data["filename"] = value["filename"]
            self.data["mimetype"] = value["mimetype"]
            self.data["encoding"] = value["encoding"]
            self.data["size"] = value["size"]
            if hasattr(value["fp"], "read"):
                self.data["fp"] = value["fp"].read()
            else:
                self.data["fp"] = value["fp"]
            return
        self.data[key] = value

    def __getitem__(self, key):
        if key == "file":
            return {
                "filename": self.data["filename"],
                "mimetype": self.data["mimetype"],
                "encoding": self.data["encoding"],
                "size": self.data["size"],
                "uid": self.uuid,
                "fp": BytesIO(self.data["fp"]),
            }
        return self.data[key]

    def __delitem__(self, key):
        if key == "file":
            del self.data["filename"]
            del self.data["mimetype"]
            del self.data["encoding"]
            del self.data["size"]
            del self.data["fp"]
        del self.data[key]

    def __dict__(self):
        res = self.data.as_dict()
        res["file"] = {
            "filename": res["filename"],
            "mimetype": res["mimetype"],
            "encoding": res["encoding"],
            "size": res["size"],
            "uid": self.uuid,
            "fp": BytesIO(res["fp"]),
        }
        for k in ["filename", "mimetype", "encoding", "size", "fp"]:
            del res[k]
        return res



class TypeBlob(Blob):
    def __init__(self, obj, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj = obj

    def open(self):
        return BytesIO(self.obj["fp"])

    def get_size(self):
        return self.obj["size"]


class TypeBlobStorage(BlobStorage):
    """ Store blob in a resource type """

    def __init__(self, request: morepath.Request, type_name: str):
        self.type_name = type_name
        self.request = request

    def put(
        self,
        fileobj: typing.BinaryIO,
        filename: str,
        mimetype: typing.Optional[str] = None,
        size: typing.Optional[int] = None,
        encoding: typing.Optional[str] = None,
        uuid: typing.Optional[str] = None,
    ) -> TypeBlob:
        data = {
            "file": {
                "uid": uuid4().hex,
                "filename": filename,
                "mimetype": mimetype,
                "size": size,
                "encoding": encoding,
                "fp": fileobj,
            }
        }
        col = self.request.get_collection(self.type_name)
        if uuid is not None:
            obj = col.get(uuid)
            obj.update(data, deserialize=False)
        else:
            obj = col.create(data, deserialize=False)

        meta = {
            "uuid": obj.uuid,
            "filename": filename,
            "mimetype": mimetype,
            "size": size,
            "encoding": encoding,
        }
        return TypeBlob(obj, **meta)

    def get(self, uuid: str) -> typing.Optional[TypeBlob]:
        col = self.request.get_collection(self.type_name)
        obj = col.get(uuid)
        if not obj:
            return None

        meta = {
            "uuid": obj.uuid,
            "filename": obj["filename"],
            "mimetype": obj["mimetype"],
            "size": obj["size"],
            "encoding": obj["encoding"],
        }
        return TypeBlob(obj, **meta)

    def delete(self, uuid: str):
        col = self.request.get_collection(self.type_name)
        obj = col.get(uuid)
        if obj:
            obj.delete()



import random
import secrets
import string
import typing
from dataclasses import dataclass, field

from morpfw.crud import Schema


def current_userid(request, data, model):
    if data["userid"] is None:
        data["userid"] = request.identity.userid
    return data["userid"]


def generate_identity(request, data, model) -> str:
    if data["api_identity"]:
        return data["api_identity"]
    return secrets.token_urlsafe(32)


@dataclass
class APIKeySchema(Schema):

    userid: typing.Optional[str] = field(
        default=None,
        metadata={"format": "uuid", "index": True, "compute_value":
            current_userid, 'editable': False, 'initializable': False},
    )
    name: typing.Optional[str] = field(default=None, metadata={"required": True})
    api_identity: typing.Optional[str] = field(
        default=None, metadata={"compute_value": generate_identity, "index":
            True, 'initializable': False, 'editable': False},
    )
    api_secret: typing.Optional[str] = field(
        default=None, metadata={"editable": False, "hidden": True,
            'initializable':False}
    )



import os
import re
import sys
import time
import warnings
from typing import Any, Dict, List, Optional

import dectate
import morepath
import reg
import sqlalchemy
import sqlalchemy.orm
import transaction
import webob
from billiard.einfo import ExceptionInfo
from celery import Celery, Task, shared_task
# from typing import Union
from celery.local import Proxy
from celery.result import AsyncResult
from celery.schedules import crontab
from more import cors
from more.transaction import TransactionApp
from morepath.reify import reify

from . import directive, exc
from .crud.app import App as CRUDApp
from .exc import ConfigurationError
from .request import DBSessionRequest, Request
from .signal.app import SignalApp


class BaseApp(CRUDApp, cors.CORSApp, SignalApp):

    home_env = 'MFW_HOME'
    request_class = Request

    @classmethod
    def all_migration_scripts(cls):
        paths = []
        for klass in cls.__mro__:
            path = getattr(klass, "migration_scripts", "migrations")
            if path:
                if not path.startswith("/"):
                    mod = sys.modules[klass.__module__]
                    filepath = getattr(mod, "__file__", None)
                    if filepath:
                        path = os.path.join(os.path.dirname(filepath), path)
                    else:
                        continue
                if os.path.exists(path):
                    if path not in paths:
                        paths.append(path)
        return paths

    def __repr__(self):
        return "<Morp Application %s:%s>" % (
            self.__class__.__module__,
            self.__class__.__name__,
        )


class App(BaseApp):
    pass


class SQLApp(TransactionApp, BaseApp):

    request_class = DBSessionRequest

    _raw_settings: dict = {}



from morpfw.crud import signals
from .app import App
from .model import PageModel


@App.subscribe(model=PageModel, signal=signals.OBJECT_CREATED)
def handle_create(request, context, signal):
    print("Object created")


@App.subscribe(model=PageModel, signal=signals.OBJECT_UPDATED)
def handle_update(request, context, signal):
    print("Object updated")


@App.subscribe(model=PageModel, signal=signals.OBJECT_TOBEDELETED)
def handle_deleted(request, context, signal):
    print("Deleting object")
    # raising exception here will prevent the deletion



import morepath
import pytz
import rulez
from more.jwtauth import JWTIdentityPolicy
from morepath import NO_IDENTITY

from ...app import SQLApp
from ...identity import Identity as BaseIdentity
from ..base import AuthnPolicy as BaseAuthnPolicy
from .apikey.model import APIKeyModel, APIKeySchema
from .app import App as BaseAuthApp
from .group.model import GroupModel, GroupSchema
from .path import get_apikey_collection, get_user_collection
from .storage.memorystorage import (
    APIKeyMemoryStorage,
    GroupMemoryStorage,
    UserMemoryStorage,
)
from .storage.sqlstorage.sqlstorage import (
    APIKeySQLStorage,
    GroupSQLStorage,
    UserSQLStorage,
)
from .user.model import UserModel, UserSchema


class Identity(BaseIdentity):
    def timezone(self):
        userid = self.userid
        collection = get_user_collection(self.request)
        user = collection.get_by_userid(userid)
        if user:
            return user.timezone()
        return pytz.UTC


class JWTWithAPIKeyIdentityPolicy(JWTIdentityPolicy):
    def identify(self, request):
        api_key = request.headers.get("X-API-KEY", None)
        if api_key:
            api_identity, api_secret = api_key.split(".")
            apikeys = get_apikey_collection(request)
            keys = apikeys.search(
                rulez.field["api_identity"] == api_identity, secure=False
            )
            if keys and keys[0].data["api_secret"] == api_secret:
                userid = keys[0].data["userid"]
                return Identity(request=request, userid=userid)
        identity = super(JWTWithAPIKeyIdentityPolicy, self).identify(request)
        if isinstance(identity, morepath.Identity):
            return Identity(request=request, userid=identity.userid)
        return identity


def verify_refresh_request(request):
    jwtauth_settings = request.app.settings.configuration.__dict__[
        "morpfw.security.jwt"
    ].copy()
    jwtauth_settings["auth_header_prefix"] = "Bearer"
    identity_policy = JWTWithAPIKeyIdentityPolicy(**jwtauth_settings)
    return identity_policy.verify_refresh(request)


class DefaultAuthnPolicy(BaseAuthnPolicy):
    def get_identity_policy(self, settings):
        jwtauth_settings = settings.configuration.__dict__["morpfw.security.jwt"]
        if jwtauth_settings:
            jwtauth_settings["auth_header_prefix"] = "Bearer"
            # Pass the settings dictionary to the identity policy.
            return JWTWithAPIKeyIdentityPolicy(**jwtauth_settings)
        raise Exception("JWTAuth configuration is not set")

    def verify_identity(self, app, identity):
        return True


@SQLApp.storage(model=UserModel)
def get_user_sqlstorage(model, request, blobstorage):
    return UserSQLStorage(request, blobstorage=blobstorage)


@SQLApp.storage(model=GroupModel)
def get_group_sqlstorage(model, request, blobstorage):
    return GroupSQLStorage(request, blobstorage=blobstorage)


@SQLApp.storage(model=APIKeyModel)
def get_apikey_sqlstorage(model, request, blobstorage):
    return APIKeySQLStorage(request, blobstorage=blobstorage)


class MemoryStorageAuthApp(BaseAuthApp):
    pass


@MemoryStorageAuthApp.storage(model=UserModel)
def get_user_memorystorage(model, request, blobstorage):
    return UserMemoryStorage(request, blobstorage=blobstorage)


@MemoryStorageAuthApp.storage(model=GroupModel)
def get_group_memorystorage(model, request, blobstorage):
    return GroupMemoryStorage(request, blobstorage=blobstorage)


@MemoryStorageAuthApp.storage(model=APIKeyModel)
def get_apikey_memorystorage(model, request, blobstorage):
    return APIKeyMemoryStorage(request, blobstorage=blobstorage)



import importlib
import json
import os
import re
import threading
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse
from uuid import uuid4

import dectate
import morepath
import transaction
from billiard.einfo import ExceptionInfo
from celery import Celery, Task, shared_task
from celery.contrib import rdb

# from typing import Union
from celery.local import Proxy
from celery.result import AsyncResult
from celery.schedules import crontab

from ..request import request_factory
from . import directive
from . import signal as event_signal

ALLOWED_ENVIRONMENT_KEYS = [
    "USER",
    "SCRIPT_NAME",
    "PATH_INFO",
    "QUERY_STRING",
    "SERVER_NAME",
    "SERVER_PORT",
    "REMOTE_HOST",
    "HTTP_HOST",
    "SERVER_PROTOCOL",
    "wsgi.url_scheme",
]


class MorpTask(Task):
    abstract = True

    def after_return(
        self,
        status: str,
        retval: Any,
        task_id: str,
        args: List,
        kwargs: Dict,
        einfo: Optional[ExceptionInfo],
    ):
        pass


class AsyncDispatcher(object):
    def __init__(self, app: morepath.App, signal: str, **kwargs):
        self.app = app
        self.signal = signal
        self.signal_opts = kwargs

    def subscribers(self) -> List[Proxy]:
        self.app.config.celery_subscriber_registry.setdefault(self.signal, [])
        return self.app.config.celery_subscriber_registry[self.signal]

    def dispatch(self, request: morepath.Request, **kwargs) -> List[AsyncResult]:
        tasks = []
        envs = {}
        for k, v in request.environ.items():
            if k in ALLOWED_ENVIRONMENT_KEYS:
                envs[k] = v

        req_json = {
            "headers": dict(request.headers),
            "environ": envs,
            # "text": request.text,
        }

        subs = self.subscribers()
        for s in subs:
            task = s.apply_async(
                kwargs=dict(request=req_json, **kwargs), **self.signal_opts
            )
            task.__signal__ = self.signal
            task.__params__ = kwargs
            self.app.dispatcher(event_signal.TASK_SUBMITTED).dispatch(request, task)
            tasks.append(task)
        return tasks


def periodic_transaction_handler(name, func):
    def transaction_wrapper(task):
        task.request.__job_name__ = name
        settings = json.loads(os.environ["MORP_SETTINGS"])
        config = settings.get("configuration", {})
        log_path = config.get("morpfw.worker.task_dir", "/tmp/")
        task_work_dir = os.path.join(
            os.path.abspath(log_path), "mfw-%s" % task.request.id
        )
        request_options = {"settings": settings, "scan": False}
        with request_factory(**request_options) as req:

            req.app.dispatcher(event_signal.SCHEDULEDTASK_STARTING).dispatch(
                req, task.request
            )

        failed = False
        try:
            cwd = os.getcwd()
            if not os.path.exists(task_work_dir):
                os.makedirs(task_work_dir)
            os.chdir(task_work_dir)
            res = func(request_options)
            os.chdir(cwd)
            with request_factory(**request_options) as req:
                req.app.dispatcher(event_signal.SCHEDULEDTASK_COMPLETED).dispatch(
                    req, task.request
                )
        except Exception:
            failed = True
            raise
        finally:
            if failed:
                with request_factory(**request_options) as req:
                    req.app.dispatcher(event_signal.SCHEDULEDTASK_FAILED).dispatch(
                        req, task.request
                    )

            with request_factory(**request_options) as req:
                req.app.dispatcher(event_signal.SCHEDULEDTASK_FINALIZED).dispatch(
                    req, task.request
                )

        return res

    return transaction_wrapper


def transaction_handler(func):
    def transaction_wrapper(task, request, **kwargs):
        settings = json.loads(os.environ["MORP_SETTINGS"])
        config = settings.get("configuration", {})
        log_path = config.get("morpfw.worker.task_dir", "/tmp/")
        task_work_dir = os.path.join(
            os.path.abspath(log_path), "mfw-%s" % task.request.id
        )
        request_options = {
            "settings": settings,
            "extra_environ": request,
            "scan": False,
        }
        with request_factory(**request_options) as req:
            req.app.dispatcher(event_signal.TASK_STARTING).dispatch(req, task.request)

        failed = False
        try:
            cwd = os.getcwd()
            if not os.path.exists(task_work_dir):
                os.makedirs(task_work_dir)
            os.chdir(task_work_dir)
            res = func(request_options, **kwargs)
            os.chdir(cwd)
            with request_factory(settings, extra_environ=request, scan=False) as req:
                req.app.dispatcher(event_signal.TASK_COMPLETED).dispatch(
                    req, task.request
                )
        except Exception:
            failed = True
            raise
        finally:
            if failed:
                with request_factory(**request_options) as req:
                    req.app.dispatcher(event_signal.TASK_FAILED).dispatch(
                        req, task.request
                    )

            with request_factory(**request_options) as req:
                req.app.dispatcher(event_signal.TASK_FINALIZED).dispatch(
                    req, task.request
                )

        return res

    return transaction_wrapper


class SignalApp(morepath.App):

    celery = Celery()
    _celery_subscribe = dectate.directive(directive.CelerySubscriberAction)

    def async_dispatcher(self, signal: str, **kwargs) -> AsyncDispatcher:
        return AsyncDispatcher(self, signal, **kwargs)

    @classmethod
    def async_subscribe(klass, signal: str, task_name: Optional[str] = None):
        def wrapper(wrapped):
            if task_name is None:
                name = ".".join([wrapped.__module__, wrapped.__name__])
            else:
                name = task_name
            func = transaction_handler(wrapped)
            task = shared_task(name=name, base=MorpTask, bind=True)(func)
            klass._celery_subscribe(signal)(task)
            return task

        return wrapper

    @classmethod
    def cron(
        klass,
        name: str,
        minute: str = "*",
        hour: str = "*",
        day_of_week: str = "*",
        day_of_month: str = "*",
        month_of_year: str = "*",
    ):
        def wrapper(wrapped):
            func = periodic_transaction_handler(name, wrapped)
            task_name = ".".join([wrapped.__module__, wrapped.__name__])
            task = shared_task(name=task_name, base=MorpTask, bind=True)(func)
            klass.celery.conf.beat_schedule[name] = {
                "task": task_name,
                "schedule": crontab(
                    minute=minute,
                    hour=hour,
                    day_of_week=day_of_week,
                    day_of_month=day_of_month,
                    month_of_year=month_of_year,
                ),
            }
            return task

        return wrapper

    @classmethod
    def periodic(klass, name: str, seconds: int = 1):
        def wrapper(wrapped):
            func = periodic_transaction_handler(name, wrapped)
            task_name = ".".join([wrapped.__module__, wrapped.__name__])
            task = shared_task(name=task_name, base=MorpTask, bind=True)(func)
            klass.celery.conf.beat_schedule[name] = {
                "task": task_name,
                "schedule": seconds,
            }
            return task

        return wrapper



import hashlib

import sqlalchemy as sa
from morpfw.crud import errors as cruderrors
from morpfw.crud.storage.sqlstorage import SQLStorage

from ... import exc
from ...apikey.model import APIKeyModel, APIKeySchema
from ...app import App
from ...group.model import GroupCollection, GroupModel, GroupSchema
from ...user.model import UserCollection, UserModel, UserSchema
from ..interfaces import IGroupStorage, IUserStorage
from . import dbmodel as db


def hash(data):
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


class UserSQLStorage(SQLStorage, IUserStorage):
    model = UserModel
    orm_model = db.User

    def create(self, collection, data):
        data["password"] = hash(data["password"])
        return super(UserSQLStorage, self).create(collection, data)

    def get_userid(self, model):
        return model.uuid

    def get_by_userid(self, collection, userid, as_model=True):
        q = self.session.query(db.User).filter(db.User.uuid == userid)
        u = q.first()
        if not u:
            return None
        if not as_model:
            return u
        return self.model(self.request, collection, u)

    def get_by_username(self, collection, username, as_model=True):
        q = self.session.query(db.User).filter(
            sa.and_(db.User.username == username, db.User.deleted.is_(None))
        )
        u = q.first()
        if not u:
            return None
        if not as_model:
            return u
        return self.model(self.request, collection, u)

    def get_by_email(self, collection, email):
        q = self.session.query(db.User).filter(
            sa.and_(db.User.email == email, db.User.deleted.is_(None))
        )
        u = q.first()
        if not u:
            return None
        return self.model(self.request, collection, u)

    def change_password(self, collection, userid, new_password):
        u = self.get_by_userid(collection, userid)
        u.data["password"] = hash(new_password)

    def get_user_groups(self, collection, userid):
        u = self.get_by_userid(collection, userid, as_model=False)
        q = (
            self.session.query(db.Membership)
            .join(db.Group)
            .filter(sa.and_(db.Membership.user_id == u.id, db.Group.deleted.is_(None)))
        )
        membership = q.all()
        groupstorage = self.request.app.get_storage(GroupModel, self.request)
        gcol = GroupCollection(self.request, groupstorage)
        res = [groupstorage.get_by_id(gcol, m.group_id) for m in membership]
        filtered = [m for m in res if not m.data["deleted"]]
        return filtered

    def validate(self, collection, userid, password):
        u = self.get_by_userid(collection, userid)
        return u.data["password"] == hash(password)

    def vacuum(self):
        roleassignment_delete = db.RoleAssignment.__table__.delete().where(
            db.User.deleted.isnot(None)
        )

        membership_delete = db.Membership.__table__.delete().where(
            db.User.deleted.isnot(None)
        )
        self.session.execute(roleassignment_delete)
        self.session.execute(membership_delete)
        return super().vacuum()


class APIKeySQLStorage(SQLStorage):
    model = APIKeyModel
    orm_model = db.APIKey


class GroupSQLStorage(SQLStorage, IGroupStorage):
    model = GroupModel
    orm_model = db.Group

    def get_user_by_userid(self, collection, userid, as_model=True):
        user_storage = self.app.get_storage(UserModel, self.request)
        user_collection = UserCollection(self.request, user_storage)
        return user_storage.get_by_userid(user_collection, userid, as_model)

    def get_user_by_username(self, collection, username, as_model=True):
        user_storage = self.app.get_storage(UserModel, self.request)
        user_collection = UserCollection(self.request, user_storage)
        return user_storage.get_by_username(user_collection, username, as_model)

    def get_members(self, collection, groupid):
        q = (
            self.session.query(db.User)
            .join(db.Membership)
            .join(db.Group)
            .filter(
                sa.and_(
                    db.Group.uuid == groupid,
                    db.Group.deleted.is_(None),
                    db.User.deleted.is_(None),
                )
            )
        )
        members = []
        user_storage = self.app.get_storage(UserModel, self.request)
        user_collection = UserCollection(self.request, user_storage)

        for m in q.all():
            members.append(user_storage.model(self.request, user_collection, m))
        return members

    def add_group_members(self, collection, groupid, userids):
        # FIXME: not using sqlalchemy relations might impact performance
        g = (
            self.session.query(db.Group)
            .filter(sa.and_(db.Group.uuid == groupid, db.Group.deleted.is_(None)))
            .first()
        )
        if not g:
            raise ValueError("Group Does Not Exist %s" % groupid)
        gid = g.id
        for userid in userids:
            u = self.get_user_by_userid(collection, userid, as_model=False)
            uid = u.id
            e = (
                self.session.query(db.Membership)
                .join(db.Group)
                .filter(
                    sa.and_(
                        db.Membership.group_id == gid,
                        db.Membership.user_id == uid,
                        db.Group.deleted.is_(None),
                    )
                )
                .first()
            )
            if not e:
                m = db.Membership()
                m.group_id = gid
                m.user_id = uid
                self.session.add(m)

    def remove_group_members(self, collection, groupid, userids):
        g = (
            self.session.query(db.Group)
            .filter(sa.and_(db.Group.uuid == groupid, db.Group.deleted.is_(None)))
            .first()
        )
        if not g:
            raise exc.GroupDoesNotExistsError(groupid)
        gid = g.id
        for userid in userids:
            u = self.get_user_by_userid(collection, userid, as_model=False)
            uid = u.id
            members = (
                self.session.query(db.Membership)
                .join(db.Group)
                .filter(
                    sa.and_(
                        db.Membership.group_id == gid,
                        db.Membership.user_id == uid,
                        db.Group.deleted.is_(None),
                    )
                )
                .all()
            )
            for m in members:
                self.session.delete(m)

    def get_group_user_roles(self, collection, groupid, userid):
        g = (
            self.session.query(db.Group)
            .filter(sa.and_(db.Group.uuid == groupid, db.Group.deleted.is_(None)))
            .first()
        )
        if not g:
            raise exc.GroupDoesNotExistsError(groupid)
        gid = g.id
        u = self.get_user_by_userid(collection, userid, as_model=False)
        uid = u.id
        roles = (
            self.session.query(db.RoleAssignment)
            .join(db.Membership)
            .join(db.Group)
            .filter(
                sa.and_(
                    db.Membership.group_id == gid,
                    db.Membership.user_id == uid,
                    db.Group.deleted.is_(None),
                )
            )
            .all()
        )
        return [r.rolename for r in roles]

    def get_by_groupname(self, collection, groupname):
        g = (
            self.session.query(db.Group)
            .filter(
                sa.and_(db.Group.groupname == groupname, db.Group.deleted.is_(None))
            )
            .first()
        )
        if not g:
            return None
        return self.model(self.request, collection, g)

    def grant_group_user_role(self, collection, groupid, userid, rolename):
        g = (
            self.session.query(db.Group)
            .filter(sa.and_(db.Group.uuid == groupid, db.Group.deleted.is_(None)))
            .first()
        )
        if not g:
            raise exc.GroupDoesNotExistsError(groupid)
        gid = g.id
        u = self.get_user_by_userid(collection, userid, as_model=False)
        uid = u.id
        m = (
            self.session.query(db.Membership)
            .join(db.Group)
            .filter(
                sa.and_(
                    db.Membership.group_id == gid,
                    db.Membership.user_id == uid,
                    db.Group.deleted.is_(None),
                )
            )
            .first()
        )
        if not m:
            raise exc.MembershipError(userid, groupid)
        ra = (
            self.session.query(db.RoleAssignment)
            .filter(
                db.RoleAssignment.membership_id == m.id,
                db.RoleAssignment.rolename == rolename,
            )
            .first()
        )
        if ra:
            return
        r = db.RoleAssignment()
        r.membership_id = m.id
        r.rolename = rolename
        self.session.add(r)

    def revoke_group_user_role(self, collection, groupid, userid, rolename):
        g = (
            self.session.query(db.Group)
            .filter(sa.and_(db.Group.uuid == groupid, db.Group.deleted.is_(None)))
            .first()
        )
        if not g:
            raise exc.GroupDoesNotExistsError(groupid)
        gid = g.id
        u = self.get_user_by_userid(collection, userid, as_model=False)
        uid = u.id
        m = (
            self.session.query(db.Membership)
            .join(db.Group)
            .filter(
                sa.and_(
                    db.Membership.group_id == gid,
                    db.Membership.user_id == uid,
                    db.Group.deleted.is_(None),
                )
            )
            .first()
        )
        if not m:
            raise exc.MembershipError(userid, groupid)
        ra = (
            self.session.query(db.RoleAssignment)
            .filter(
                db.RoleAssignment.membership_id == m.id,
                db.RoleAssignment.rolename == rolename,
            )
            .first()
        )
        if ra:
            self.session.delete(ra)

    def vacuum(self):
        roleassignment_delete = db.RoleAssignment.__table__.delete().where(
            db.Group.deleted.isnot(None)
        )
        membership_delete = db.Membership.__table__.delete().where(
            db.Group.deleted.isnot(None)
        )
        self.session.execute(roleassignment_delete)
        self.session.execute(membership_delete)
        return super().vacuum()



import copy
from ..blobstorage.base import NullBlobStorage
from ..errors import BlobStorageNotImplementedError
from ...interfaces import IStorage


class BaseStorage(IStorage):

    use_transactions = True

    blobstorage = None

    @property
    def model(self):
        raise NotImplementedError

    def __init__(self, request, blobstorage=None):
        self.request = request
        self.app = request.app
        self.blobstorage = blobstorage
        super().__init__(request, blobstorage)

    def set_identifier(self, obj, identifier):
        idfield = self.app.get_identifierfield(self.model.schema)
        obj[idfield] = identifier

    def set_schema_defaults(self, data):
        obj = self.model.schema(**data)
        return obj.__dict__

#	def create(self, data):
#		raise NotImplementedError
#
#	def search(self, query=None, limit=None):
#		raise NotImplementedError
#
#	def aggregate(self, query=None, group=None, order_by=None):
#		raise NotImplementedError
#
#	def get(self, identifier):
#		raise NotImplementedError
#
#	def get_by_id(self, id):
#		raise NotImplementedError
#
#	def get_by_uuid(self, uuid):
#		raise NotImplementedError
#
#	def update(self, identifier, data):
#		raise NotImplementedError
#
#	def delete(self, identifier, model):
#		raise NotImplementedError

    def _blob_guard(self):
        if self.blobstorage is None or isinstance(self.blobstorage, NullBlobStorage):
            raise BlobStorageNotImplementedError(
                'Storage does not implement blobstorage')

    def get_blob(self, uuid):
        self._blob_guard()
        return self.blobstorage.get(uuid)

    def put_blob(self, fileobj, filename, mimetype=None, size=None, encoding=None):
        self._blob_guard()
        blob = self.blobstorage.put(
            fileobj, filename, mimetype, size, encoding)
        return blob

    def delete_blob(self, uuid):
        self._blob_guard()
        self.blobstorage.delete(uuid)



import typing
from dataclasses import dataclass
from morpfw.crud.xattrprovider.fieldxattrprovider import FieldXattrProvider
import morpfw
from .app import App
from .model import PageModel


@dataclass
class PageXattrSchema(morpfw.BaseSchema):

    field1: typing.Optional[str] = None
    field2: typing.Optional[str] = None


class PageXattrProvider(FieldXattrProvider):

    schema = PageXattrSchema
    additional_properties = False


@App.xattrprovider(model=PageModel)
def get_xattrprovider(context):
    return PageXattrProvider(context)



from morepath import NO_IDENTITY

from ..identity import Identity
from .base import AuthnPolicy as BaseAuthnPolicy


class NoAuthIdentityPolicy(object):
    def identify(self, request):
        return Identity(request=request, userid=None)

    def remember(self, response, request, identity):
        pass

    def forget(self, response, request):
        pass


class AuthnPolicy(BaseAuthnPolicy):
    def get_identity_policy(self, settings):
        return NoAuthIdentityPolicy()

    def verify_identity(self, app, identity):
        return True



from dataclasses import dataclass, field

from morpfw import Collection, Model, Schema
from morpfw.app import BaseApp
from morpfw.authz.pas import DefaultAuthzPolicy

from .test_auth import login, logout


class App(BaseApp, DefaultAuthzPolicy):
    pass


@dataclass
class Object1Schema(Schema):

    title: str = field(
        default="", metadata={"format": "text"},
    )


class Object1Collection(Collection):
    schema = Object1Schema


class Object1Model(Model):
    schema = Object1Schema


@App.path(model=Object1Collection, path="object1")
def object1_collection_factory(request):
    storage = request.app.get_storage(Object1Model, request)
    return Object1Collection(request, storage)


@App.path(model=Object1Model, path="object1/{identifier}")
def object1_model_factory(request, identifier):
    collection = object1_collection_factory(request)
    return collection.get(identifier)


@App.typeinfo(name="tests.object1", schema=Object1Schema)
def get_object1_typeinfo(request):
    return {
        "title": "Object1",
        "description": "Object1 type",
        "schema": Object1Schema,
        "collection": Object1Collection,
        "collection_factory": object1_collection_factory,
        "model": Object1Model,
        "model_factory": object1_model_factory,
    }


@dataclass
class Object2Schema(Schema):

    title: str = field(
        default="", metadata={"format": "text"},
    )


class Object2Collection(Collection):
    schema = Object2Schema


class Object2Model(Model):
    schema = Object2Schema


@App.path(model=Object2Collection, path="object2")
def object2_collection_factory(request):
    storage = request.app.get_storage(Object2Model, request)
    return Object2Collection(request, storage)


@App.path(model=Object2Model, path="object2/{identifier}")
def object2_model_factory(request, identifier):
    collection = object2_collection_factory(request)
    return collection.get(identifier)


@App.typeinfo(name="tests.object2", schema=Object2Schema)
def get_object2_typeinfo(request):
    return {
        "title": "Object2",
        "description": "Object2 type",
        "schema": Object2Schema,
        "collection": Object2Collection,
        "collection_factory": object2_collection_factory,
        "model": Object2Model,
        "model_factory": object2_model_factory,
    }


def run_test(c):

    login(c, "admin")

    r = c.post_json(
        "/user/+register",
        {
            "username": "user1",
            "email": "user1@localhost.com",
            "timezone": "UTC",
            "password": "password",
            "password_validate": "password",
        },
    )

    login(c, "user1")

    r = c.get("/object1/+search")

    r = c.post_json("/object1/", {"title": "Hello"})

    obj_uuid = r.json["data"]["uuid"]

    r = c.get("/object1/%s" % obj_uuid)

    r = c.patch_json("/object1/%s" % obj_uuid, {"title": "World"})

    r = c.get("/object1/%s" % obj_uuid)

    r = c.delete("/object1/%s" % obj_uuid)

    r = c.get("/object2/")

    r = c.get("/object2/+search")

    r = c.post_json("/object2/", {"title": "Hello"}, expect_errors=True)

    assert r.status_code == 403

    login(c, "admin")

    r = c.post_json("/object2/", {"title": "Hello"})

    obj2_uuid = r.json["data"]["uuid"]

    login(c, "user1")

    r = c.get("/object2/%s" % obj2_uuid, expect_errors=True)

    assert r.status_code == 403

    r = c.patch_json("/object2/%s" % obj2_uuid, {"title": "World"}, expect_errors=True)

    assert r.status_code == 403

    r = c.delete("/object2/%s" % obj2_uuid, expect_errors=True)

    assert r.status_code == 403



import inspect
from uuid import uuid4

import dectate
import reg

from . import pubsub

OBJECT_CREATED = "morpfw.object_created"
OBJECT_UPDATED = "morpfw.object_updated"
OBJECT_TOBEDELETED = "morpfw.object_tobedeleted"


def _get_identifier(obj):
    if inspect.isclass(obj):
        return [obj.__module__, obj.__name__]
    if getattr(obj, "__class__", None):
        return [obj.__module__, obj.__class__.__name__]
    raise ValueError("%s is neither instance nor class" % obj)


def construct_key(signal, app, obj):
    return ":".join([signal] + _get_identifier(obj))


class Connect(dectate.Action):

    app_class_arg = True

    def __init__(self, model, signal):
        self.model = model
        self.signal = signal

    def identifier(self, app_class):
        return uuid4().hex

    def perform(self, obj, app_class):
        app_class._events.subscribe(obj, model=self.model, signal=self.signal)


class Dispatcher(object):
    def __init__(self, app, signal):
        self.app = app
        self.signal = signal

    def dispatch(self, request, context):
        self.app._events.publish(self.app, request, context, self.signal)


class SignalApp(dectate.App):

    subscribe = dectate.directive(Connect)

    @reg.dispatch_method(
        reg.match_instance("model", lambda self, request, obj, signal: obj),
        reg.match_key("signal", lambda self, request, obj, signal: signal),
    )
    def _events(self, request, obj, signal):
        raise NotImplementedError

    def dispatcher(self, signal):
        return Dispatcher(self, signal)



import morepath
from more.jwtauth import JWTIdentityPolicy
from morpfw.crud import App as CRUDApp
import reg
import dectate
from morepath.reify import reify
import json
from typing import List, Optional, Type
import morpfw
from ...app import BaseApp

App = BaseApp



import typing
from dataclasses import dataclass, field

import morpfw
import morpfw.sql
import sqlalchemy as sa


@dataclass
class PageSchema(morpfw.Schema):

    title: typing.Optional[str] = field(default=None, metadata={"title": "Title"})
    body: typing.Optional[str] = field(default=None, metadata={"title": "Body"})


class PageCollection(morpfw.Collection):
    schema = PageSchema


class PageModel(morpfw.Model):
    schema = PageSchema


# SQLALchemy model
class Page(morpfw.sql.Base):

    __tablename__ = "test_page"

    title = sa.Column(sa.String(length=1024))
    body = sa.Column(sa.Text())


class PageStorage(morpfw.SQLStorage):
    model = PageModel
    orm_model = Page


@App.storage(model=PageModel)
def get_storage(model, request, blobstorage):
    return PageStorage(request, blobstorage=blobstorage)


@App.path(model=PageCollection, path="/pages")
def get_collection(request):
    storage = request.app.get_storage(PageModel, request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path="/pages/{identifier}")
def get_model(request, identifier):
    col = get_collection(request)
    return col.get(identifier)


@App.permission_rule(model=PageCollection, permission=All)
def allow_collection_all(identity, context, permission):
    """ Default permission rule, allow all """
    return True


@App.permission_rule(model=PageModel, permission=All)
def allow_model_all(identity, context, permission):
    """ Default permission rule, allow all """
    return True


@App.typeinfo(name="test.page", schema=PageSchema)
def get_typeinfo(request):
    return {
        "title": "Test Page",
        "description": "",
        "schema": PageSchema,
        "collection": PageCollection,
        "collection_factory": get_collection,
        "model": PageModel,
        "model_factory": get_model,
    }



import morpfw

settings = {
    "application": {
        "title": "My App",  # app title
        "class": "app:App",  # import path to your app
    }
}

with morpfw.request_factory(settings) as request:
    # do something here with the request
    pass



from copy import copy, deepcopy
from dataclasses import field

MISSING = object()


class Field(object):
    def __init__(self, metadata=None):
        self.metadata = metadata or {}
        self.metadata.setdefault("required", False)
        self.metadata.setdefault("editable", True)
        self._default = MISSING
        self._default_factory = MISSING

    def default(self, value):
        self._default = value
        return self

    def default_factory(self, value):
        self._default_factory = value
        return self

    def required(self, required=True):
        self.metadata["required"] = required
        return self

    def editable(self, editable=True):
        self.metadata["editable"] = editable
        return self

    def init(self):
        if self._default is not MISSING:
            return field(default=self._default, metadata=self.metadata)
        if self._default_factory is not MISSING:
            return field(default_factory=self._default_factory, metadata=self.metadata)
        return field(metadata=self.metadata)

    def widget(self, widget):
        self.metadata["deform.widget"] = widget
        return self

    def unique(self):
        self.metadata["unique"] = True
        return self



import secrets
import typing
from dataclasses import dataclass, field

import deform.widget
import pytz
from morpfw.crud.schema import BaseSchema, Schema
from morpfw.crud.validator import regex_validator

from ..app import App
from ..model import EMAIL_PATTERN, NAME_PATTERN
from .validator import valid_source, valid_tz


@dataclass
class RegistrationSchema(BaseSchema):
    # pattern=NAME_PATTERN)
    username: typing.Optional[str] = field(
        default=None,
        metadata={
            "required": True,
            "validators": [regex_validator(NAME_PATTERN, "name")],
        },
    )

    email: typing.Optional[str] = field(
        default=None,
        metadata={
            "required": True,
            "validators": [regex_validator(EMAIL_PATTERN, "email")],
        },
    )

    password: typing.Optional[str] = field(default=None, metadata={"required": True})
    password_validate: typing.Optional[str] = field(
        default=None, metadata={"required": True}
    )


@dataclass
class LoginSchema(BaseSchema):
    username: typing.Optional[str] = field(default=None, metadata={"required": True})

    password: typing.Optional[str] = field(default=None, metadata={"required": True})


@dataclass
class UserSchema(Schema):

    username: typing.Optional[str] = field(
        default=None,
        metadata={
            "required": True,
            "editable": False,
            "searchable": True,
            "validators": [regex_validator(NAME_PATTERN, "name")],
        },
    )

    email: typing.Optional[str] = field(
        default=None,
        metadata={
            "required": True,
            "searchable": True,
            "validators": [regex_validator(EMAIL_PATTERN, "email")],
        },
    )

    password: typing.Optional[str] = field(
        default=None,
        metadata={
            "required": True,
            "editable": False,
            "hidden": True,
            "deform.widget": deform.widget.PasswordWidget(),
        },
    )
    nonce: typing.Optional[str] = field(
        default_factory=lambda: secrets.token_hex(8),
        metadata={"deform.widget": deform.widget.HiddenWidget(), "hidden": True},
    )
    source: typing.Optional[str] = field(
        default="local", metadata={"validators": [valid_source]}
    )
    timezone: typing.Optional[str] = field(
        default="UTC",
        metadata={
            "validators": [valid_tz],
            "required": True,
            "deform.widget": deform.widget.Select2Widget(
                values=[(x, x) for x in pytz.all_timezones]
            ),
        },
    )
    is_administrator: typing.Optional[bool] = field(
        default=False, metadata={"searchable": True}
    )

    __unique_constraint__ = ["username", "deleted"]



import os
import time

import pytest
from more.basicauth import BasicAuthIdentityPolicy
from morpfw.app import SQLApp
from morpfw.request import request_factory

from .common import get_client, start_worker


class App(SQLApp):
    pass


class Root(object):
    pass


def get_identity_policy():
    return BasicAuthIdentityPolicy


def verify_identity(identity):
    return True


@App.path(model=Root, path="")
def get_root(request):
    return Root()


@App.json(model=Root)
def index(context, request):
    subs = request.app.async_dispatcher("test_signal").dispatch(
        request, obj={"data": 10}
    )
    cel = request.app.celery
    res = []
    for s in subs:
        try:
            res.append(s.get())
        except:
            pass
    return res


@App.async_subscribe("test_signal")
def handler1(request_options, obj):
    request_options["scan"] = False
    with request_factory(**request_options) as request:
        obj["handler"] = "handler1"
        obj["data"] += 1
    return obj


@App.async_subscribe("test_signal")
def handler2(request_options, obj):
    request_options["scan"] = False
    with request_factory(**request_options) as request:
        obj["handler"] = "handler2"
        obj["data"] += 5
    return obj


@App.async_subscribe("test_signal")
def handler3(request_options, obj):
    request_options["scan"] = False
    with request_factory(**request_options) as request:
        obj["handler"] = "handler3"
    raise Exception("Error")


def test_signal(pgsql_db, pika_connection_channel, celery_worker):
    config = os.path.join(os.path.dirname(__file__), "test_worker-settings.yml")
    c = get_client(
        config, get_identity_policy=get_identity_policy, verify_identity=verify_identity
    )

    c.authorization = ("Basic", ("dummy", "dummy"))

    r = c.get("/")

    res = list(sorted(r.json, key=lambda x: x["handler"]))
    assert res[0]["data"] == 11
    assert res[1]["data"] == 15



import cProfile
import os
import runpy
import sys
import time

import click
from alembic.config import main as alembic_main

from . import db, elasticstore, register_admin, scheduler, shell, start, worker
from .generate_config import genconfig


def run_module(argv=sys.argv):
    if len(argv) <= 1:
        print("Usage: %s [module]" % argv[0])
        sys.exit(1)
    mod = argv[1]
    sys.argv = argv[0:1] + argv[2:]
    runpy.run_module(mod, run_name="__main__", alter_sys=True)


def run_module_profile(argv=sys.argv):
    mod = argv[1]
    sys.argv = argv[0:1] + argv[2:]
    prof = cProfile.Profile()
    starttime = time.time()
    prof.enable()
    runpy.run_module(mod, run_name="__main__", alter_sys=True)
    prof.disable()
    endtime = time.time()
    print(f"Time taken: {endtime - starttime:.3f} seconds")
    outfile = mod + ".pstats"
    if os.path.exists(outfile):
        os.unlink(outfile)
    prof.dump_stats(outfile)
    print(f"Profiler result stored as {outfile}")


if __name__ == "__main__":
    main()



import jsl
import copy


def _avro_field(**kwargs):
    result = {}
    for k, v in kwargs.items():
        if v:
            result[k] = v
    return result


def jsl_field_to_avro_field(prop: jsl.BaseField, name, namespace) -> dict:
    if isinstance(prop, jsl.DateTimeField):
        return _avro_field(name=name, type='string')
    if isinstance(prop, jsl.StringField):
        return _avro_field(name=name, type='string')
    if isinstance(prop, jsl.IntField):
        return _avro_field(name=name, type='int')
    if isinstance(prop, jsl.DictField):
        return _avro_field(name=name, type='map')
    if isinstance(prop, jsl.NumberField):
        return _avro_field(name=name, type='double')
    if isinstance(prop, jsl.BooleanField):
        return _avro_field(name=name, type='boolean')
    if isinstance(prop, jsl.DocumentField):
        if prop.document_cls:
            subtype = jsl_to_avro(
                prop.document_cls, name=name, namespace='%s.%s' % (namespace, name))
            return _avro_field(name=name, type=subtype)
        return _avro_field(name=name, type='map')
    if isinstance(prop, jsl.ArrayField):
        if prop.items:
            if isinstance(prop.items, jsl.DocumentField):
                subtype = jsl_to_avro(
                    prop.items.document_cls, namespace='%s.%s' % (namespace, name))
            elif isinstance(prop.items, jsl.BaseField):
                subtype = jsl_field_to_avro_field(
                    prop.items, name=prop.name, namespace=namespace)['type']
            else:
                raise KeyError(prop.items)
            return _avro_field(name=name, type='array', items=subtype)
        return _avro_field(name=name, type='array', items='string')

    raise KeyError(prop)


def jsl_to_avro(schema, name=None, namespace='schema'):
    fields = []
    avsc = {
        'type': 'record',
    }
    if name:
        avsc['name'] = name
    if namespace:
        avsc['namespace'] = namespace
    for attr, prop in schema._fields.items():
        prop.name = attr
        fields.append(jsl_field_to_avro_field(
            prop, name=attr, namespace=namespace))

    avsc['fields'] = fields
    return avsc



import typing

from transitions import Machine

from ...interfaces import IStateMachine


class StateMachine(IStateMachine):

    readonly_states: list = []
    transition_guards: dict = {}  # key-value, transition_name->permission

    def __init__(self, context, **kwargs):
        self._context = context
        self._request = context.request
        self._app = context.request.app
        initial = self.state or self.states[0]
        kwargs.update(
            {
                "model": self,
                "transitions": self.transitions,
                "states": self.states,
                "initial": initial,
                "auto_transitions": False,
            }
        )
        self._machine = Machine(**kwargs)

    def _get_state(self) -> typing.Optional[str]:
        try:
            return self._context.data["state"]
        except KeyError:
            return None

    def _set_state(self, val):
        self._context.data["state"] = val

    state = property(_get_state, _set_state)

    def get_triggers(self):
        context = self._context
        request = self._request
        triggers = [
            i for i in self._machine.get_triggers(self.state) if not i.startswith("to_")
        ]
        result = []
        for trigger in triggers:
            guard = self.transition_guards.get(trigger, None)
            if not guard:
                result.append(trigger)
            else:
                if request.permits(context, guard):
                    result.append(trigger)
        return result

    def is_readonly(self):
        if self.state in self.readonly_states:
            return True
        return False




import pytz

from ....crud.errors import FieldValidationError


def valid_source(request, schema, field, value, mode=None):
    if value not in ["local", "ldap"]:
        return "Only local & ldap source is supported at the moment"


def valid_tz(request, schema, field, value, mode=None):
    if value not in pytz.all_timezones:
        return "Invalid timezone"



import copy
import importlib
import os
import sys
import tempfile
from configparser import RawConfigParser

import click
import morpfw
import reg
from alembic.config import CommandLine as AlembicCLI
from alembic.config import Config as AlembicCfg
from alembic.config import main as alembic_main

from ..alembic import drop_all
from .cli import cli, confirmation_dialog, load


@reg.dispatch(reg.match_class("app", lambda app_cls, settings: app_cls))
def alembic_config(app_cls, settings: dict):
    raise NotImplementedError("alembic_config for %s" % app_cls)


@reg.dispatch(reg.match_class("app", lambda app_cls, settings: app_cls))
def alembic_logging_config(app_cls, settings: dict) -> dict:
    raise NotImplementedError("alembic_logging_config %s" % app_cls)


@alembic_logging_config.register(app=morpfw.BaseApp)
def default_alembic_logging_config(app_cls, settings: dict):
    return {
        "version": 1,
        "loggers": {
            "": {"level": "WARN", "handlers": ["console"], "qualname": ""},
            "sqlalchemy": {"level": "WARN", "qualname": "sqlalchemy.engine",},
            "alembic": {"level": "INFO", "qualname": "alembic"},
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
                "level": "NOTSET",
                "formatter": "generic",
            }
        },
        "formatters": {
            "generic": {
                "format": r"%(levelname)-5.5s [%(name)s] %(message)s",
                "datefmt": r"%H:%M:%S",
            }
        },
    }


@alembic_config.register(app=morpfw.BaseApp)
def default_alembic_config(app_cls, settings: dict) -> AlembicCfg:
    config = settings["configuration"]
    proj = "default"
    acfg = AlembicCfg()
    script_location = config.get(
        "alembic.migration_script_location", os.path.join(os.getcwd(), "migrations")
    )
    acfg.set_main_option("script_location", script_location)
    acfg.set_main_option("databases", proj)
    acfg.set_section_option(
        proj, "sqlalchemy.url", config["morpfw.storage.sqlstorage.dburl"]
    )
    logging_cfg = alembic_logging_config(app_cls, settings)

    acfg.logging_config = logging_cfg
    return acfg


@cli.command(help="Reset database")
@click.pass_context
def resetdb(ctx):
    param = load(ctx.obj["settings"])

    settings = param["settings"]

    if not confirmation_dialog():
        return

    with morpfw.request_factory(settings) as request:
        drop_all(request)


@cli.command(help="Vacuum database")
@click.pass_context
def vacuum(ctx):
    param = load(ctx.obj["settings"])

    settings = param["settings"]

    with morpfw.request_factory(settings) as request:

        types = request.app.config.type_registry.get_typeinfos(request)
        for typeinfo in types.values():

            collection = request.get_collection(typeinfo["name"])
            vacuum_f = getattr(collection.storage, "vacuum", None)
            if vacuum_f:
                print("Vacuuming %s" % typeinfo["name"])
                items = vacuum_f()
                print("%s record(s) affected" % items)


@cli.command(
    help="manage alembic migration",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_context
def db(ctx):
    param = load(ctx.obj["settings"])
    settings = param["settings"]

    modname, cname = settings["application"]["class"].split(":")
    mod = importlib.import_module(modname)
    app_cls = getattr(mod, cname)

    acfg = alembic_config(app_cls, settings)

    argv = copy.copy(sys.argv)
    if "-s" in argv:
        sindex = argv.index("-s")
        argv = argv[:sindex] + argv[sindex + 2 :]
    if "--settings" in argv:
        sindex = argv.index("--settings")
        argv = argv[:sindex] + argv[sindex + 2 :]

    prog = "%s %s" % (argv[0], argv[1])
    acli = AlembicCLI(prog=prog)
    options = acli.parser.parse_args(argv[2:])
    if not hasattr(options, "cmd"):
        acli.parser.error("too few arguments")
    else:
        acli.run_cmd(acfg, options)



import typing
import uuid
from datetime import datetime
from decimal import Decimal

import jsl
import sqlalchemy as sa
import sqlalchemy.orm as saorm
import sqlalchemy_jsonfield as sajson
from inverter import dc2pgsqla
from rulez import compile_condition
from sqlalchemy import Column, Index, MetaData, Table, func
from sqlalchemy import types as satypes
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.exc import StatementError
from sqlalchemy.types import CHAR, TypeDecorator

from ..app import App
from .base import BaseStorage
from .sqlstorage import Base, MappedTable, SQLStorage

db_meta = Base.metadata

_mapped_models = {}


def construct_orm_model(schema, metadata, *, name=None):
    existing = _mapped_models.get(schema, None)
    if existing:
        return existing

    table = dc2pgsqla.convert(schema, metadata, name=name)

    class Table(MappedTable):

        __table__ = table

    saorm.mapper(Table, table)

    _mapped_models[schema] = Table
    return Table


class PgSQLStorage(SQLStorage):
    """
    At the moment this storage only works correctly with EntityContent.

    To use with code level SQLStorage, use `construct_orm_model`
    """

    _temp: dict = {}
    __table_name__ = None

    def __init__(self, request, metadata=None, blobstorage=None):
        super().__init__(request, blobstorage=blobstorage)
        self.metadata = metadata or db_meta
        self.metadata.bind = self.session.get_bind()

    @property
    def orm_model(self):
        name = getattr(self, "__table_name__", None)
        return construct_orm_model(self.model.schema, self.metadata, name=name)



import base64
import functools
import secrets

import morepath
from jwt import InvalidTokenError
from more.jwtauth import JWTIdentityPolicy
from oauthlib.common import Request as OAuthRequest
from oauthlib.oauth2 import RequestValidator as BaseRequestValidator
from oauthlib.oauth2 import Server
from oauthlib.oauth2.rfc6749 import errors
from webob.exc import HTTPBadRequest, HTTPForbidden

import morpfw
from morpfw.authn.base import AuthnPolicy as BaseAuthnPolicy
from morpfw.authn.pas.policy import Identity
from morpfw.crud.util import resolve_model

from .authn.pas.app import App


def jwtpolicy(request: morpfw.Request) -> JWTIdentityPolicy:
    settings = request.app.settings.configuration.__dict__["morpfw.security.jwt"].copy()
    return JWTIdentityPolicy(**settings)


class RequestValidator(BaseRequestValidator):
    def __init__(self, request: morpfw.Request) -> None:
        self.mfw_request = request

        super().__init__()

    def validate_client_id(self, client_id: str, request: OAuthRequest):
        clients = self.mfw_request.get_collection("morpfw.pas.apikey")
        if clients.get_by_identity(client_id):
            return True
        return False

    def authenticate_client(self, request, *args, **kwargs):
        clients = self.mfw_request.get_collection("morpfw.pas.apikey")
        client = None
        if request._params["refresh_token"]:
            jwt = jwtpolicy(self.mfw_request)
            tokendata = jwt.decode_jwt(request._params["refresh_token"])
            client = clients.get_by_identity(tokendata["client_id"])
            request.client = client
        else:
            auth_method, auth_token = request.headers["Authorization"].split(" ")
            client_id, client_secret = (
                base64.b64decode(auth_token.encode("utf-8")).decode("utf-8").split(":")
            )
            client = clients.get_by_identity(client_id)
            if client.validate(client_secret):
                request.client = client
        if client is not None:
            return True
        return False

    def validate_grant_type(
        self, client_id, grant_type, client, request, *args, **kwargs
    ):
        if grant_type in ["client_credentials", "authorization_code", "refresh_token"]:
            return True
        return False

    def get_default_scopes(self, client_id, request, *args, **kwargs):
        return ["home", "admin"]

    def validate_scopes(self, client_id, scopes, client, request, *args, **kwargs):
        return True

    def save_bearer_token(self, token, request, *args, **kwargs):
        return self.mfw_request.relative_url("/")

    #    def validate_code(self, client_id, code, client, request, *args, **kwargs):
    #        # FIXME
    #        return True
    #
    #    def validate_refresh_token(self, refresh_token, client, request, *args, **kwargs):
    #        jwt = jwtpolicy(self.mfw_request)
    #        try:
    #            tokendata = jwt.decode_jwt(request._params["refresh_token"])
    #        except InvalidTokenError:
    #            return False
    #        return True
    #
    #    def get_original_scopes(self, refresh_token, request, *args, **kwargs):
    #        return self.get_default_scopes(request.client.client_id, request)
    #
    #    def revoke_token(self, token, token_type_hint, request, *args, **kwargs):
    #        jwt = jwtpolicy(self.mfw_request)
    #        data = jwt.decode_jwt(token, False)
    #        users = self.mfw_request.get_collection("morpfw.pas.user")
    #        user = users.get_by_userid(data["userid"])
    #        if user:
    #            user.update({"nonce": secrets.token_hex(8)})
    #        return
    #
    #    def save_authorization_code(self, client_id, code, request, *args, **kwargs):
    #        # FIXME
    #        return
    #
    #    def validate_redirect_uri(self, client_id, redirect_uri, request, *args, **kwargs):
    #        # FIXME
    #        return True
    #
    #    def confirm_redirect_uri(
    #        self, client_id, code, redirect_uri, client, request, *args, **kwargs
    #    ):
    #        # FIXME
    #        return True
    #
    #    def invalidate_authorization_code(self, client_id, code, request, *args, **kwargs):
    #        # FIXME
    #        return
    #
    #
    #    def validate_response_type(
    #        self, client_id, response_type, client, request, *args, **kwargs
    #    ):
    #        if response_type in ["code"]:
    #            return True
    #        return False

    def generate_access_token(self, request):
        jwt = jwtpolicy(self.mfw_request)
        user = request.client.user()
        identity = Identity(
            self.mfw_request, user.userid, client_id=request.client.client_id
        )
        claims = identity.as_dict()
        userid = claims.pop("userid")
        claims_set = jwt.create_claims_set(self.mfw_request, userid, claims)
        token = jwt.encode_jwt(claims_set)
        return token

    def get_default_redirect_uri(self, client_id, request, *args, **kwargs):
        return self.mfw_request.relative_url("/")


def oauthserver(request: morpfw.Request) -> Server:
    validator = RequestValidator(request)
    return Server(
        validator,
        token_generator=validator.generate_access_token,
        refresh_token_generator=validator.generate_access_token,
    )


class OAuthRoot(object):
    def __init__(self, request, server_factory=oauthserver) -> None:
        self.request = request
        self.server = server_factory(request)


def extract_params(request: morepath.Request):
    uri = request.url
    http_method = request.method
    body = request.body
    headers = dict(request.headers)
    return uri, http_method, body, headers


class OAuthProvider(object):
    def __init__(self, context, request) -> None:
        self.context = context
        self.request = request
        self.server = context.server

    def authorization_form(self):
        request = self.request
        uri, http_method, body, headers = extract_params(request)

        try:
            (scopes, credentials,) = self.server.validate_authorization_request(
                uri, http_method, body, headers
            )
            # Not necessarily in session but they need to be
            # accessible in the POST view after form submit.
            request.session["oauth2_credentials"] = credentials
            request.session.save()
            return self._authorization_form(scopes, credentials)

        # Errors that should be shown to the user on the provider website
        except errors.FatalClientError as e:
            return self.response_from_error(e)

        # Errors embedded in the redirect URI back to the client
        except errors.OAuth2Error as e:
            return morepath.redirect(e.in_uri(e.redirect_uri))

    def _authorization_form(self, scopes, credentials):
        credentials = credentials.copy()
        del credentials["request"]
        return {
            "scopes": scopes,
            "credentials": credentials,
            "submit_url": "/authorize",
        }

    def authorize(self):
        request = self.request
        uri, http_method, body, headers = extract_params(request)

        # The scopes the user actually authorized, i.e. checkboxes
        # that were selected.
        scopes = request.POST.getall("scopes")

        # Extra credentials we need in the validator
        # tem
        credentials = {"user": request.identity.userid}

        # The previously stored (in authorization GET view) credentials
        credentials.update(request.session.get("oauth2_credentials", {}))
        try:
            headers, body, status = self.server.create_authorization_response(
                uri, http_method, body, headers, scopes, credentials
            )
            return self.response_from_return(headers, body, status)

        except errors.FatalClientError as e:
            return self.response_from_error(e)

    def get_token(self):
        request = self.request
        uri, http_method, body, headers = extract_params(request)
        # If you wish to include request specific extra credentials for
        # use in the validator, do so here.
        credentials = {}

        headers, body, status = self.server.create_token_response(
            uri, http_method, body, headers, credentials
        )

        # All requests to /token will return a json response, no redirection.
        return self.response_from_return(headers, body, status)

    def refresh_token(self):
        request = self.request
        uri, http_method, body, headers = extract_params(request)

        headers, body, status = self.server.create_revocation_response(
            uri, http_method, body, headers
        )

        # All requests to /token will return a json response, no redirection.
        return self.response_from_return(headers, body, status)

    def response_from_return(self, headers, body, status):
        return morepath.Response(body=body, headers=headers, status=status)

    def response_from_error(self, e):
        raise HTTPBadRequest()


# @App.json(model=OAuthRoot, name="authorize")
# def authorize(context, request):
#    oauth = OAuthProvider(context, request)
#    res = oauth.authorization_form()
#    return res
#
#
# @App.json(model=OAuthRoot, name="authorize", request_method="POST")
# def authorize(context, request):
#    oauth = OAuthProvider(context, request)
#    res = oauth.authorize()
#    return res
#


@App.json(model=OAuthRoot, name="token", request_method="POST")
def get_token(context, request):
    oauth = OAuthProvider(context, request)
    res = oauth.get_token()
    return res


class AuthnPolicy(BaseAuthnPolicy):
    def get_identity_policy(self, settings):
        jwtauth_settings = settings.configuration.__dict__["morpfw.security.jwt"]
        jwtauth_settings["auth_header_prefix"] = "Bearer"
        if jwtauth_settings:
            # Pass the settings dictionary to the identity policy.
            return JWTIdentityPolicy(**jwtauth_settings)
        raise Exception("JWTAuth configuration is not set")

    def verify_identity(self, app, identity):
        return True


def hook_oauth_models(cls, path="/oauth2"):
    @cls.path(model=OAuthRoot, path=path)
    def get_authroot(request):
        return OAuthRoot(request)


App.hook_oauth_models = classmethod(hook_oauth_models)



from ...crud.errors import AlreadyExistsError


class AuthManagerError(Exception):
    pass


class InvalidPasswordError(AuthManagerError):
    pass


class UserExistsError(AuthManagerError, AlreadyExistsError):
    pass


class GroupExistsError(AuthManagerError, AlreadyExistsError):
    pass


class GroupDoesNotExistsError(AuthManagerError):
    pass


class UserDoesNotExistsError(AuthManagerError):
    pass


class MembershipError(AuthManagerError):
    def __init__(self, user, group, *args, **kwargs):
        msg = "User %s not in group %s" % (user, group)
        super(MembershipError, self).__init__(msg, *args, **kwargs)


class InvalidRoleError(AuthManagerError):
    pass



import re

from rulez import OperatorNotAllowedError, parse_dsl

from ...interfaces import IAggregateProvider
from ..errors import UnprocessableError
from ..model import Collection


class AggregateProvider(IAggregateProvider):

    field_pattern = re.compile(r"(\w+):(\w+)")
    function_pattern = re.compile(r"(\w+):([\_\w]+)\((\w+)\)")

    def __init__(self, context: Collection):
        self.context = context
        self.storage = context.storage
        self.request = context.request

    def _parse(self, qs):
        tokens = [ss.strip() for ss in qs.strip().split(",")]
        result = []
        for t in tokens:
            m = self.function_pattern.match(t)
            if m:
                g = m.groups()
                result.append((g[0], {"function": g[1], "field": g[2]}))
                continue
            m = self.field_pattern.match(t)
            if m:
                f, v = t.split(":")
                result.append((f, v))
                continue
            raise ValueError(t)

        return dict(result)

    def parse_query(self, qs):
        if not qs.strip():
            return None
        try:
            return parse_dsl(qs)
        except ValueError:
            raise UnprocessableError("Invalid search query '%s'" % qs)
        except OperatorNotAllowedError:
            raise UnprocessableError("Invalid search query '%s'" % qs)
        return None

    def parse_group(self, qs):
        if not qs.strip():
            return None
        try:
            return self._parse(qs)
        except ValueError:
            raise UnprocessableError("Invalid aggregate query '%s'" % qs)
        return None



from morpfw.crud import pubsub
import reg


def test_signal():

    class Model(object):
        pass

    class SubModel(Model):
        pass

    @reg.dispatch(reg.match_instance('model'),
                  reg.match_key('signal', lambda model, signal: signal))
    def event(model, signal):
        raise NotImplementedError

    @event.subscribe(model=Model, signal='event')
    def one(model, signal):
        return 1

    @event.subscribe(model=Model, signal='event')
    def two(model, signal):
        return 2

    @event.subscribe(model=SubModel, signal='event')
    def three(model, signal):
        return 3

    mobj = Model()
    smobj = SubModel()
    assert list(sorted(event.publish(model=mobj, signal='event'))) == [1, 2]
    assert list(
        sorted(event.publish(model=smobj, signal='event'))) == [1, 2, 3]


def test_event_dispatchmethod():

    class App(object):

        @reg.dispatch_method(reg.match_instance('model'),
                             reg.match_key('signal', lambda self, model, signal: signal))
        def event(self, model, signal):
            raise NotImplementedError

    class Model(object):
        pass

    class SubModel(Model):
        pass

    @App.event.subscribe(model=Model, signal='event')
    def one(app, model, signal):
        return 1

    @App.event.subscribe(model=Model, signal='event')
    def two(app, model, signal):
        return 2

    @App.event.subscribe(model=SubModel, signal='event')
    def three(app, model, signal):
        return 3

    app = App()
    mobj = Model()
    smobj = SubModel()
    assert list(sorted(
        app.event.publish(app, model=mobj, signal='event'))) == [1, 2]
    assert list(sorted(
        app.event.publish(app, model=smobj, signal='event'))) == [1, 2, 3]



import click
import morpfw

from .cli import cli, confirmation_dialog, load


@cli.command(help="Update elasticsearch indexes")
@click.pass_context
def update_esindex(ctx):
    param = load(ctx.obj["settings"])

    settings = param["settings"]

    with morpfw.request_factory(settings) as request:

        types = request.app.config.type_registry.get_typeinfos(request)
        for typeinfo in types.values():
            collection = request.get_collection(typeinfo["name"])
            storage = collection.storage
            if isinstance(storage, morpfw.ElasticSearchStorage):
                print("Creating index %s .. " % storage.index_name, end="")
                if storage.create_index(collection):
                    print("OK")
                else:
                    if storage.update_index(collection):
                        print("UPDATED")
                    else:
                        raise AssertionError("Unknown error")


@cli.command(help="delete all elasticsearch indexes")
@click.pass_context
def reset_esindex(ctx):

    if not confirmation_dialog():
        return

    param = load(ctx.obj["settings"])

    settings = param["settings"]

    with morpfw.request_factory(settings) as request:

        types = request.app.config.type_registry.get_typeinfos(request)
        client = request.get_es_client()
        for typeinfo in types.values():
            collection = request.get_collection(typeinfo["name"])
            storage = collection.storage
            if isinstance(storage, morpfw.ElasticSearchStorage):
                print("Deleting index %s .. " % storage.index_name, end="")
                client.indices.delete(storage.index_name)
                print("OK")




import typing
import uuid
from datetime import datetime
from decimal import Decimal

import jsl
import pytz
import sqlalchemy as sa
import sqlalchemy_jsonfield as sajson
import sqlalchemy_utils as sautils
from rulez import compile_condition
from sqlalchemy import func
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.exc import StatementError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql import select
from sqlalchemy.types import CHAR, TypeDecorator
from zope.sqlalchemy import mark_changed

from ..app import App
from .base import BaseStorage


class SQLStorage(BaseStorage):

    _temp: dict = {}

    @property
    def orm_model(self):
        raise NotImplementedError

    @property
    def session(self):
        return self.request.db_session

    def create(self, collection, data):
        o = self.orm_model()
        dst = self.app.get_dataprovider(self.model.schema, o, self)
        src = self.app.get_dataprovider(self.model.schema, data, self)
        for k, v in src.items():
            dst[k] = v
        m = self.model(self.request, collection, o)
        identifier = m.identifier
        self.session.add(o)
        self.session.flush()
        self.session.refresh(o)
        return m

    def aggregate(self, query=None, group=None, order_by=None, limit=None):
        group_bys = []
        group_bys_map = {}

        if group:
            fields = []
            for k, v in group.items():
                if isinstance(v, str):
                    c = getattr(self.orm_model, v)
                    fields.append(c)
                    group_bys.append(c)
                elif isinstance(v, dict):
                    ff = v["function"]
                    f = v["field"]
                    c = getattr(self.orm_model, f)
                    if ff == "count":
                        op = func.count(c).label(k)
                        fields.append(op)
                        group_bys_map[k] = op
                    elif ff == "sum":
                        op = func.sum(c).label(k)
                        fields.append(op)
                        group_bys_map[k] = op
                    elif ff == "avg":
                        op = func.avg(c).label(k)
                        fields.append(op)
                        group_bys_map[k] = op
                    elif ff == "min":
                        op = func.min(c).label(k)
                        fields.append(op)
                        group_bys_map[k] = op
                    elif ff == "max":
                        op = func.max(c).label(k)
                        fields.append(op)
                        group_bys_map[k] = op
                    elif ff == "year":
                        op = func.date_part("YEAR", c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                        group_bys_map[k] = op
                    elif ff == "month":
                        op = func.date_part("MONTH", c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                        group_bys_map[k] = op
                    elif ff == "day":
                        op = func.date_part("DAY", c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                        group_bys_map[k] = op
                    elif ff == "date":
                        op = func.to_char(c, "YYYY-MM-DD").label(k)
                        fields.append(op)
                        group_bys.append(op)
                        group_bys_map[k] = op
                    elif ff == "hourly":
                        op = func.to_char(c, "YYYY-MM-DD HH24:00 TZ").label(k)
                        fields.append(op)
                        group_bys.append(op)
                        group_bys_map[k] = op
                    else:
                        raise ValueError("Unknown function %s" % ff)
        else:
            fields = [self.orm_model]

        include_deleted = self.request.environ.get(
            "morpfw.sqlstorage.include_deleted", False
        )
        if query:
            f = compile_condition("sqlalchemy", query)
            filterquery = f(self.orm_model)
            if not include_deleted:
                filterquery = sa.and_(self.orm_model.deleted.is_(None), filterquery)
            q = self.session.query(*fields).filter(filterquery)
        else:
            q = self.session.query(*fields).filter(self.orm_model.deleted.is_(None))

        if order_by is not None:
            col = order_by[0]
            d = order_by[1]
            if d not in ["asc", "desc"]:
                raise KeyError(d)
            if col in group_bys_map:
                colattr = group_bys_map[col]
            else:
                colattr = getattr(self.orm_model, col)
            if d == "desc":
                q = q.order_by(colattr.desc())
            else:
                q = q.order_by(colattr)

        if group_bys:
            q = q.group_by(*group_bys)

        if limit is not None:
            q = q.limit(limit)

        results = []
        try:
            q_res = q.all()
        except StatementError:
            empty_res = {}
            for f in fields:
                empty_res[f.key] = 0
            return [empty_res]

        for o in q_res:
            d = o._asdict()
            for k, v in d.items():
                if isinstance(v, datetime):
                    d[k] = v.isoformat()
                elif isinstance(v, Decimal):
                    d[k] = float(v)
                elif isinstance(v, uuid.UUID):
                    d[k] = v.hex

            results.append(d)
        return results

    def search(self, collection, query=None, offset=None, limit=None, order_by=None):
        include_deleted = self.request.environ.get(
            "morpfw.sqlstorage.include_deleted", False
        )
        if query:
            f = compile_condition("sqlalchemy", query)
            filterquery = f(self.orm_model)
            if not include_deleted:
                q = self.session.query(self.orm_model).filter(
                    sa.and_(self.orm_model.deleted.is_(None), filterquery)
                )
            else:
                q = self.session.query(self.orm_model).filter(filterquery)
        else:
            if not include_deleted:
                q = self.session.query(self.orm_model).filter(
                    self.orm_model.deleted.is_(None)
                )
            else:
                q = self.session.query(self.orm_model)

        if order_by is not None:
            col = order_by[0]
            d = order_by[1]
            if d not in ["asc", "desc"]:
                raise KeyError(d)
            colattr = getattr(self.orm_model, col)
            if d == "desc":
                q = q.order_by(colattr.desc())
            else:
                q = q.order_by(colattr)
        if offset is not None:
            q = q.offset(offset)
        if limit is not None:
            q = q.limit(limit)

        yield_per = self.request.environ.get("morpfw.sqlstorage.yield_per", None)
        if yield_per is None:
            try:
                return [self.model(self.request, collection, o) for o in q.all()]
            except StatementError:
                return []
        else:
            try:
                return [
                    self.model(self.request, collection, o)
                    for o in q.yield_per(yield_per)
                ]
            except StatementError:
                return []

    def get(self, collection, identifier):
        qs = []
        idfield = self.app.get_identifierfield(self.model.schema)
        include_deleted = self.request.environ.get(
            "morpfw.sqlstorage.include_deleted", False
        )

        if not include_deleted:
            q = self.session.query(self.orm_model).filter(
                sa.and_(
                    self.orm_model.deleted.is_(None),
                    getattr(self.orm_model, idfield) == identifier,
                )
            )
        else:
            q = self.session.query(self.orm_model).filter(
                getattr(self.orm_model, idfield) == identifier,
            )
        try:
            r = q.first()
        except StatementError:
            return None
        if not r:
            return None
        return self.model(self.request, collection, r)

    def get_by_id(self, collection, id):
        q = self.session.query(self.orm_model).filter(self.orm_model.id == id)
        r = q.first()
        if not r:
            return None
        return self.model(self.request, collection, r)

    def get_by_uuid(self, collection, uuid):
        uuid_field = self.app.get_uuidfield(self.model.schema)
        if getattr(self.orm_model, uuid_field, None) is None:
            raise AttributeError(
                "%s does not have %s field" % (self.orm_model, uuid_field)
            )
        qs = getattr(self.orm_model, uuid_field) == uuid
        q = self.session.query(self.orm_model).filter(qs)
        r = q.first()
        if not r:
            return None
        return self.model(self.request, collection, r)

    def update(self, collection, identifier, data):
        qs = []

        idfield = self.app.get_identifierfield(self.model.schema)
        include_deleted = self.request.environ.get(
            "morpfw.sqlstorage.include_deleted", False
        )
        qs.append(getattr(self.orm_model, idfield) == identifier)
        if not include_deleted:
            qs.append(self.orm_model.deleted.is_(None))
        q = self.session.query(self.orm_model).filter(sa.and_(*qs))

        r = q.first()
        if not r:
            raise ValueError(identifier)

        d = self.app.get_dataprovider(self.model.schema, r, self)
        for k, v in data.items():
            if d.get(k, None) != v:
                d[k] = v
        self.session.flush()
        return self.model(self.request, collection, r)

    def delete(self, identifier, model, **kwargs):
        permanent = kwargs.get("permanent", False)
        if permanent:
            model.delete()
            return

        qs = []
        idfield = self.app.get_identifierfield(self.model.schema)
        qs.append(getattr(self.orm_model, idfield) == identifier)
        q = self.session.query(self.orm_model).filter(sa.and_(*qs))

        r = q.first()
        if not r:
            raise ValueError(identifier)

        d = self.app.get_dataprovider(self.model.schema, r, self)
        d["deleted"] = datetime.now(tz=pytz.UTC)

    def vacuum(self):
        affected = select([func.count(self.orm_model.__table__.c.uuid)]).where(
            self.orm_model.deleted.isnot(None)
        )
        delete_q = self.orm_model.__table__.delete().where(
            self.orm_model.deleted.isnot(None)
        )
        items = self.session.execute(affected)
        total = items.fetchone()[0]
        self.session.execute(delete_q)
        mark_changed(self.session())
        return total


GUID = sautils.UUIDType


class MappedTable(object):
    pass


class BaseMixin(MappedTable):

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(GUID, default=uuid.uuid4, index=True, unique=True)
    created = sa.Column(
        sa.DateTime(timezone=True),
        default=lambda: datetime.now(tz=pytz.UTC),
        index=True,
    )
    creator = sa.Column(GUID, index=True)
    modified = sa.Column(
        sa.DateTime(timezone=True),
        default=lambda: datetime.now(tz=pytz.UTC),
        index=True,
    )
    state = sa.Column(sa.String(length=64), index=True)
    deleted = sa.Column(sa.DateTime(timezone=True), index=True)
    blobs = sa.Column(sajson.JSONField)
    xattrs = sa.Column(sajson.JSONField)


Base = declarative_base(cls=BaseMixin)



from morepath import NO_IDENTITY
from ..identity import Identity
from .base import AuthnPolicy as BaseAuthnPolicy
from ipaddress import IPv4Address, IPv4Network


class UserParamIdentityPolicy(object):

    def __init__(self, allowed_nets=None):
        self.allowed_nets = []
        for n in (allowed_nets or []):
            self.allowed_nets.append(IPv4Network(n))

    def identify(self, request):
        userid = request.GET.get('user.id', None)
        if userid:
            remote_addr = IPv4Address(request.remote_addr)
            for n in self.allowed_nets:
                if remote_addr in n:
                    return Identity(userid)
        return NO_IDENTITY

    def remember(self, response, request, identity):
        pass

    def forget(self, response, request):
        pass


class AuthnPolicy(BaseAuthnPolicy):

    @classmethod
    def get_identity_policy(cls, settings):
        return UserParamIdentityPolicy(allowed_nets=settings.security.allowed_nets)

    @classmethod
    def verify_identity(cls, app, identity):
        return True



from more.itsdangerous import IdentityPolicy
from .base import AuthnPolicy as BaseAuthnPolicy


class AuthnPolicy(BaseAuthnPolicy):

    def get_identity_policy(self, settings):
        if settings.application.development_mode:
            secure = False
        else:
            secure = True
        return IdentityPolicy(secure=secure)

    def verify_identity(self, app, identity):
        return True



import os

import jsl
import morpfw.crud.signals as signals
from more.basicauth import BasicAuthIdentityPolicy
from more.transaction import TransactionApp
from morpfw.crud.blobstorage.fsblobstorage import FSBlobStorage
from morpfw.crud.storage.memorystorage import MemoryStorage

from ..common import get_client
from .crud_common import FSBLOB_DIR
from .crud_common import App as BaseApp
from .crud_common import (
    BlobObjectCollection,
    BlobObjectModel,
    NamedObjectCollection,
    NamedObjectModel,
    ObjectCollection,
    ObjectModel,
    PageCollection,
    PageModel,
    PageSchema,
    run_jslcrud_test,
)


class App(BaseApp):
    pass


class PageStorage(MemoryStorage):
    model = PageModel


@App.path(model=PageCollection, path="pages")
def collection_factory(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path="pages/{identifier}")
def model_factory(request, identifier):
    col = collection_factory(request)
    return col.get(identifier)


@App.typeinfo(name="tests.page", schema=PageSchema)
def get_page_typeinfo(request):
    return {
        "title": "Page",
        "description": "Page type",
        "schema": PageSchema,
        "collection": PageCollection,
        "collection_factory": collection_factory,
        "model": PageModel,
        "model_factory": model_factory,
    }


class ObjectStorage(MemoryStorage):
    incremental_id = True
    incremental_column = "id"
    model = ObjectModel


@App.path(model=ObjectCollection, path="objects")
def object_collection_factory(request):
    storage = ObjectStorage(request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path="objects/{identifier}")
def object_model_factory(request, identifier):
    col = object_collection_factory(request)
    return col.get(identifier)


class NamedObjectStorage(MemoryStorage):
    model = NamedObjectModel


@App.path(model=NamedObjectCollection, path="named_objects")
def namedobject_collection_factory(request):
    storage = NamedObjectStorage(request)
    return NamedObjectCollection(request, storage)


@App.path(model=NamedObjectModel, path="named_objects/{identifier}")
def namedobject_model_factory(request, identifier):
    col = namedobject_collection_factory(request)
    return col.get(identifier)


class BlobObjectStorage(MemoryStorage):
    model = BlobObjectModel


@App.storage(model=BlobObjectModel)
def get_blobobject_storage(model, request, blobstorage):
    return BlobObjectStorage(request, blobstorage=blobstorage)


@App.blobstorage(model=BlobObjectModel)
def get_blobobject_blobstorage(model, request):
    return FSBlobStorage(request, FSBLOB_DIR)


@App.path(model=BlobObjectCollection, path="blob_objects")
def blobobject_collection_factory(request):
    storage = request.app.get_storage(BlobObjectModel, request)
    return BlobObjectCollection(request, storage)


@App.path(model=BlobObjectModel, path="blob_objects/{identifier}")
def blobobject_model_factory(request, identifier):
    col = blobobject_collection_factory(request)
    return col.get(identifier)


def test_memorystorage():
    config = os.path.join(os.path.dirname(__file__), "test_memorystorage-settings.yml")
    client = get_client(config)
    run_jslcrud_test(client, skip_aggregate=True)



import copy
import importlib
import multiprocessing
import os
import subprocess
import tempfile

import morepath
import reg
import sqlalchemy
import transaction
import yaml
from beaker.middleware import CacheMiddleware as BeakerCacheMiddleware
from beaker.middleware import SessionMiddleware as BeakerMiddleware
from celery import Celery
from more.basicauth import BasicAuthIdentityPolicy
from zope.sqlalchemy import register as register_session

from .app import BaseApp, SQLApp
from .authn.pas.group.model import GroupModel, GroupSchema
from .authn.pas.group.path import get_group_collection
from .authn.pas.user.model import UserCollection, UserModel, UserSchema
from .authn.pas.user.path import get_user_collection
from .exc import ConfigurationError
from .request import Request, default_settings, request_factory
from .sql import Base


def create_app(settings, scan=True, **kwargs):
    assert "class" in settings["application"]
    app_mod, app_clsname = settings["application"]["class"].split(":")
    app = getattr(importlib.import_module(app_mod), app_clsname)

    s = copy.deepcopy(default_settings)
    for k in settings.keys():
        if k in s.keys():
            for j, v in settings[k].items():
                s[k][j] = v
        else:
            s[k] = settings[k]

    settings = s

    # initialize app
    config = settings["configuration"]

    if scan:
        morepath.autoscan()
        for scanmodpath in config["morpfw.scan"] or []:
            scanmod = importlib.import_module(scanmodpath)
            morepath.scan(package=scanmod)

    authnpolicy_settings = config["morpfw.authn.policy.settings"]
    authnpol_mod, authnpol_clsname = config["morpfw.authn.policy"].strip().split(":")

    authnpolicy = getattr(importlib.import_module(authnpol_mod), authnpol_clsname)(
        authnpolicy_settings
    )

    get_identity_policy = authnpolicy.get_identity_policy
    verify_identity = authnpolicy.verify_identity

    app.identity_policy()(get_identity_policy)
    app.verify_identity()(verify_identity)
    app.init_settings(settings)
    app._raw_settings = settings

    if config["app.development_mode"]:
        os.environ["MOREPATH_TEMPLATE_AUTO_RELOAD"] = "1"

    if not app.is_committed():
        app.commit()

    celery_settings = config["morpfw.celery"]
    app.celery.conf.update(**celery_settings)
    application = app()

    # wrap with beaker session and cache manager

    beaker_settings = {}

    for k, v in settings["configuration"].items():
        if k.startswith("morpfw.beaker."):
            keylen = len("morpfw.beaker.")
            beaker_settings[k[keylen:]] = v

    if "session.type" in beaker_settings:
        application = BeakerMiddleware(application, beaker_settings)
    if "cache.type" in beaker_settings:
        application = BeakerCacheMiddleware(application, beaker_settings)
    return application


def create_admin(request: Request, username: str, password: str, email: str):
    transaction.manager.begin()
    usercol = get_user_collection(request)
    userobj = usercol.create(
        {
            "username": username,
            "password": password,
            "email": email,
            "state": "active",
            "source": "local",
            "is_administrator": True,
            "timezone": "UTC",
        }
    )
    gcol = request.get_collection("morpfw.pas.group")
    group = gcol.get_by_groupname("__default__")
    group.add_members([userobj.userid])
    group.grant_member_role(userobj.userid, "administrator")
    transaction.manager.commit()
    return userobj


def run(app, settings, host="127.0.0.1", port=5000, ignore_cli=True):
    morepath.run(app, host=host, port=port, ignore_cli=ignore_cli)


def runprod(app, settings, host="127.0.0.1", port=5000, ignore_cli=True, workers=None):
    service = "gunicorn"
    server_listen = {"listen_address": host, "listen_port": port}
    server = settings["server"].copy()
    server.update(server_listen)
    opts = {}
    opts["loglevel"] = server.get("log_level", "INFO")
    opts["log_directory"] = settings.get("logging", {}).get("log_directory", "/tmp")
    os.environ["MORP_APP_FACTORY"] = settings["application"]["factory"]
    logconfig = (
        """
[loggers]
keys=root, gunicorn.error, gunicorn.access

[handlers]
keys=console, error_file, access_file, application_file

[formatters]
keys=generic, access

[logger_root]
level=%(loglevel)s
handlers=console, application_file

[logger_gunicorn.error]
level=%(loglevel)s
handlers=error_file
propagate=1
qualname=gunicorn.error

[logger_gunicorn.access]
level=%(loglevel)s
handlers=access_file
propagate=0
qualname=gunicorn.access

[handler_console]
class=StreamHandler
formatter=generic
args=(sys.stdout, )

[handler_error_file]
class=logging.FileHandler
formatter=generic
args=('%(log_directory)s/errors.log',)

[handler_access_file]
class=logging.FileHandler
formatter=access
args=('%(log_directory)s/access.log',)

[handler_application_file]
class=logging.FileHandler
formatter=generic
args=('%(log_directory)s/application.log',)


[formatter_generic]
format=%%(asctime)s [%%(process)d] [%%(levelname)s] %%(message)s
datefmt=%%Y-%%m-%%d %%H:%%M:%%S
class=logging.Formatter

[formatter_access]
format=%%(message)s
class=logging.Formatter
    """
        % opts
    )
    if workers is None:
        workers = (multiprocessing.cpu_count() * 2) + 1

    logconf = tempfile.mktemp()
    with open(logconf, "w") as f:
        f.write(logconfig)

    opts = [
        "--log-config",
        logconf,
        "-b",
        "%(listen_address)s:%(listen_port)s" % server,
        "--workers",
        str(server.get("workers", workers)),
        "--max-requests",
        str(server.get("max_requests", 1000)),
        "--max-requests-jitter",
        str(server.get("max_requests_jitter", 1000)),
        "--worker-connections",
        str(server.get("worker_connections", 1000)),
        "--timeout",
        str(server.get("worker_timeout", 30)),
    ]

    subprocess.call([service] + opts + ["morpfw.wsgi:app"])


def set_buildout_environ(config: str) -> None:
    envs = {}
    for l in config.strip().split("\n"):
        sep = l.find(":")
        if sep < 0:
            raise ValueError(l)
        k = l[:sep].strip()
        v = l[sep + 1 :].strip()
        envs[k] = v

    for k, v in envs.items():
        os.environ[k] = v

    if "BUILDOUT_BINDIR" in envs.keys():
        bindir = envs["BUILDOUT_BINDIR"]
        os.environ["PATH"] = "%s:%s" % (bindir, os.environ["PATH"])




from datetime import datetime
from uuid import uuid4

import pytz

from . import model, signals
from .app import App


@App.subscribe(signal=signals.OBJECT_CREATED, model=model.Model)
def set_uuid(app, request, obj, signal):
    uuid_field = app.get_uuidfield(obj.schema)
    if uuid_field in obj.schema.__dataclass_fields__.keys():
        if obj.data.get(uuid_field, None) is None:
            obj.data[uuid_field] = uuid4().hex


@App.subscribe(signal=signals.OBJECT_CREATED, model=model.Model)
def set_created(app, request, obj, signal):
    if "created" in obj.schema.__dataclass_fields__.keys():
        now = datetime.now(tz=pytz.UTC)
        obj.data["created"] = now
        obj.data["modified"] = now


@App.subscribe(signal=signals.OBJECT_CREATED, model=model.Model)
def set_creator(app, request, obj, signal):
    if "creator" in obj.schema.__dataclass_fields__.keys():
        obj.data["creator"] = request.identity.userid or None


@App.subscribe(signal=signals.OBJECT_UPDATED, model=model.Model)
def set_modified(app, request, obj, signal):
    if "modified" in obj.schema.__dataclass_fields__.keys():
        obj.data["modified"] = datetime.now(tz=pytz.UTC)



import json
import os
import shutil
import tempfile
import typing
from dataclasses import dataclass, field
from datetime import date, datetime
from uuid import uuid4

import jsl
import morepath
import morpfw.crud.signals as signals
import pytz
import yaml
from more.basicauth import BasicAuthIdentityPolicy
from morpfw.app import BaseApp
from morpfw.authn.base import AuthnPolicy as BaseAuthnPolicy
from morpfw.crud import permission as crudperm
from morpfw.crud.model import Collection, Model
from morpfw.crud.schema import BaseSchema, Schema
from morpfw.crud.statemachine.base import StateMachine
from morpfw.crud.xattrprovider import FieldXattrProvider
from morpfw.crud.xattrprovider.base import XattrProvider
from morpfw.identity import Identity
from morpfw.interfaces import ISchema
from morpfw.main import create_app
from webtest import TestApp as Client

FSBLOB_DIR = tempfile.mkdtemp()

epoch = date(1970, 1, 1)


class App(BaseApp):
    pass


@App.permission_rule(model=Model, permission=crudperm.All)
def allow_all_model_access(identity, context, permission):
    return True


@App.permission_rule(model=Model, permission=crudperm.StateUpdate)
def allow_all_model_state_access(identity, context, permission):
    return True


@App.permission_rule(model=Collection, permission=crudperm.All)
def allow_all_collection_access(identity, context, permission):
    return True


def validate_body(request, schema, data, mode=None, **kw):
    if not isinstance(data["body"], str):
        return

    if data["body"].lower() == "invalid":
        return {"field": "body", "message": "Body must not be 'invalid'"}


@dataclass
class PageSchema(Schema):

    title: str = field(
        default="", metadata={"format": "text",},
    )
    body: str = ""
    publish_date: typing.Optional[date] = None
    publish_datetime: typing.Optional[datetime] = None
    value: typing.Optional[int] = None
    footer: typing.Optional[str] = ""
    computed_title: typing.Optional[str] = field(
        default=None, metadata={"compute_value": lambda req, d, m: "computed title"}
    )


@App.formvalidators(schema=PageSchema)
def page_formvalidators(schema):
    return [validate_body]


class PageCollection(Collection):
    schema = PageSchema


class PageModel(Model):
    schema = PageSchema


@dataclass
class ObjectSchema(Schema):

    body: str = ""
    created_flag: typing.Optional[bool] = False
    updated_flag: typing.Optional[bool] = False
    attrs: typing.Optional[dict] = field(default_factory=dict)


class ObjectCollection(Collection):
    schema = ObjectSchema


class ObjectModel(Model):
    schema = ObjectSchema


@App.subscribe(signal=signals.OBJECT_CREATED, model=ObjectModel)
def object_created(app, request, obj, signal):
    obj.data["created_flag"] = True


@App.subscribe(signal=signals.OBJECT_UPDATED, model=ObjectModel)
def object_updated(app, request, obj, signal):
    obj.data["updated_flag"] = True


@dataclass
class ObjectXattrSchema(BaseSchema):

    message: typing.Optional[str] = None
    message_date: typing.Optional[date] = None


class ObjectXattrProvider(FieldXattrProvider):

    schema = ObjectXattrSchema


@App.xattrprovider(model=ObjectModel)
def get_objectmodel_xattrprovider(context):
    return ObjectXattrProvider(context)


class PageStateMachine(StateMachine):

    states = ["new", "pending", "approved"]
    transitions = [
        {"trigger": "approve", "source": ["new", "pending"], "dest": "approved"},
        {"trigger": "submit", "source": "new", "dest": "pending"},
    ]


@App.statemachine(model=PageModel)
def get_pagemodel_statemachine(context):
    return PageStateMachine(context)


@dataclass
class NamedObjectSchema(Schema):

    name: typing.Optional[str] = None
    body: typing.Optional[str] = ""
    created_flag: typing.Optional[bool] = False
    updated_flag: typing.Optional[bool] = False


@App.identifierfield(schema=NamedObjectSchema)
def namedobject_identifierfield(schema):
    return "name"


@App.default_identifier(schema=NamedObjectSchema)
def namedobject_default_identifier(schema, obj, request):
    return obj["name"]


class NamedObjectCollection(Collection):
    schema = NamedObjectSchema


@App.json(model=NamedObjectCollection, name="get_uuid")
def get_object_by_uuid(context, request):
    uuid = request.GET.get("uuid")
    return context.get_by_uuid(uuid).json()


class NamedObjectModel(Model):
    schema = NamedObjectSchema


class BlobObjectSchema(Schema):
    pass


class BlobObjectCollection(Collection):
    schema = BlobObjectSchema


class BlobObjectModel(Model):
    schema = BlobObjectSchema

    blob_fields = ["blobfile"]


_USERUIDS = {}


class IdentityPolicy(BasicAuthIdentityPolicy):
    def identify(self, request):
        identity = super().identify(request)
        if isinstance(identity, morepath.Identity):
            uuid = _USERUIDS.get(identity.userid, None)
            if not uuid:
                uuid = uuid4().hex
                _USERUIDS[identity.userid] = uuid

            return Identity(
                request=request,
                userid=uuid,
                username=identity.userid,
                password=identity.password,
            )
        return identity


class AuthnPolicy(BaseAuthnPolicy):
    def get_identity_policy(self, settings):
        return IdentityPolicy()

    def verify_identity(self, app, identity):
        if identity.username == "admin" and identity.password == "admin":
            return True
        return False


def run_jslcrud_test(c, skip_aggregate=False):

    c.authorization = ("Basic", ("admin", "admin"))

    # test loading the model details and schema
    r = c.get("/pages")

    assert r.json["schema"]["type"] == "object"

    if not skip_aggregate:
        # test simple count
        r = c.get("/pages/+aggregate", {"group": ("count:count(uuid)")},)
        assert r.json[0]["count"] == 0

    # lets try creating an entry
    publish_date = (date(2019, 1, 1) - epoch).days
    publish_datetime = int(
        datetime(2019, 1, 1, 1, 1, tzinfo=pytz.UTC).timestamp() * 1000
    )
    r = c.post_json(
        "/pages/",
        {
            "title": "Hello",
            "body": "World",
            "publish_date": publish_date,
            "publish_datetime": publish_datetime,
        },
    )

    assert r.json["data"]["title"] == "Hello"
    assert r.json["data"]["publish_date"] == publish_date
    assert r.json["data"]["publish_datetime"] == publish_datetime
    assert r.json["data"]["computed_title"] == "computed title"
    uuid = r.json["data"]["uuid"]
    assert uuid
    assert len(uuid) == 32

    r = c.get("/pages/%s" % uuid)

    assert r.json["data"]["title"] == "Hello"
    assert r.json["data"]["state"] == "new"

    # lets see if the entry appears in listing

    for i in range(10):
        r = c.post_json(
            "/pages/", {"title": "page%s" % i, "body": "page%sbody" % i, "value": i}
        )

    if not skip_aggregate:
        r = c.get(
            "/pages/+aggregate",
            {
                "group": (
                    "count:count(uuid), year:year(created), month:month(created),"
                    "day:day(created), sum:sum(value), avg:avg(value)"
                )
            },
        )

        now = datetime.utcnow()
        assert r.json[0]["year"] == now.year
        assert r.json[0]["month"] == now.month
        assert r.json[0]["day"] == now.day
        assert r.json[0]["sum"] == 45
        assert r.json[0]["count"] == 11
        assert r.json[0]["avg"] == 4.5

        r = c.get("/pages/+aggregate", {"group": ("created:created"), "limit": 1},)

        assert len(r.json) == 1

    r = c.get("/pages/+search", {"q": 'title in ["Hello","something"]'})

    assert r.json["results"][0]["data"]["title"] == "Hello"

    r = c.get("/pages/+search")

    assert "Hello" in [i["data"]["title"] for i in r.json["results"]]

    r = c.get("/pages/+search", {"select": "$.title", "q": 'title in ["Hello"]'})

    assert r.json["results"] == [["Hello"]]

    r = c.get(
        "/pages/+search", {"select": "$.[title, body]", "q": 'title in ["Hello"]'}
    )

    assert r.json["results"] == [["Hello", "World"]]

    r = c.get("/pages/+search", {"order_by": "title"})

    assert list([i["data"]["title"] for i in r.json["results"]]) == (
        ["Hello"] + ["page%s" % i for i in range(10)]
    )

    r = c.get("/pages/+search", {"order_by": "title:desc"})

    assert list([i["data"]["title"] for i in r.json["results"]]) == (
        ["page%s" % i for i in range(9, -1, -1)] + ["Hello"]
    )

    r = c.get("/pages/+search", {"order_by": "title", "limit": 5})

    assert list([i["data"]["title"] for i in r.json["results"]]) == (
        ["Hello"] + ["page%s" % i for i in range(4)]
    )

    r = c.get("/pages/+search", {"order_by": "title", "offset": 1})

    assert list([i["data"]["title"] for i in r.json["results"]]) == (
        ["page%s" % i for i in range(10)]
    )

    r = c.get("/pages/+search", {"order_by": "title", "offset": 1, "limit": 5})

    assert list([i["data"]["title"] for i in r.json["results"]]) == (
        ["page%s" % i for i in range(5)]
    )

    # lets create another with wrong invalid values
    r = c.post_json(
        "/pages/", {"title": "page2", "body": 123, "footer": 123}, expect_errors=True
    )

    assert r.json["status"] == "error"
    assert len(r.json["field_errors"]) == 2
    assert len(r.json["form_errors"]) == 0

    r = c.post_json(
        "/pages/",
        {"title": "page2", "body": "valid", "footer": 123},
        expect_errors=True,
    )

    assert r.json["status"] == "error"
    assert len(r.json["field_errors"]) == 1

    r = c.post_json(
        "/pages/",
        {"title": "page2", "body": "invalid", "footer": "footer"},
        expect_errors=True,
    )

    assert r.json["status"] == "error"
    assert len(r.json["field_errors"]) == 1

    # lets update the entry
    r = c.patch_json("/pages/%s" % uuid, {"body": "newbody"})

    assert r.json["status"] == "success"

    r = c.get("/pages/%s" % uuid)

    assert r.json["data"]["body"] == "newbody"

    # lets approve the page

    r = c.post_json("/pages/%s/+statemachine" % uuid, {"transition": "approve"})

    r = c.get("/pages/%s" % uuid)

    assert r.json["data"]["state"] == "approved"

    # it cant be approved twice

    r = c.post_json(
        "/pages/%s/+statemachine" % uuid, {"transition": "approve"}, expect_errors=True
    )

    assert r.status_code == 422

    r = c.patch_json("/pages/%s" % uuid, {"body": "invalid"}, expect_errors=True)

    assert r.json["status"] == "error"

    # lets delete the entry

    r = c.get("/pages/+search")

    assert len(r.json["results"]) == 11

    r = c.delete("/pages/%s" % uuid)

    assert r.json["status"] == "success"

    r = c.get("/pages/%s" % uuid, expect_errors=True)

    r = c.get("/pages/+search")

    assert len(r.json["results"]) == 10

    r = c.get("/objects")

    assert r.json

    r = c.post_json("/objects/", {"body": "hello"})

    assert r.json["data"]["body"] == "hello"
    assert r.json["data"]["created_flag"] is True
    assert r.json["data"]["created"]
    assert r.json["data"]["creator"] == _USERUIDS["admin"]

    obj_link = r.json["links"][0]["href"]
    obj_xattr_link = obj_link + "/+xattr"
    r = c.get(obj_xattr_link)

    assert r.json == {}

    r = c.get(obj_link)

    assert (
        r.json["data"].get("xattrs", None) is None
        or r.json["data"].get("xattrs", None) == {}
    )

    r = c.get(obj_link + "/+xattr-schema")

    assert r.json["schema"]["$schema"]

    r = c.patch_json(
        obj_xattr_link,
        {"message": "hello world", "message_date": "201"},
        expect_errors=True,
    )

    assert r.status_code == 422

    message_date = (date(2019, 1, 1) - epoch).days
    r = c.patch_json(
        obj_xattr_link, {"message": "hello world", "message_date": message_date},
    )

    assert r.status_code == 200

    r = c.get(obj_xattr_link)

    assert r.json == {"message": "hello world", "message_date": message_date}

    r = c.get(obj_link)
    assert r.json["data"]["xattrs"] == {
        "message": "hello world",
        "message_date": message_date,
    }

    r = c.patch_json(obj_xattr_link, {"message": "hello world", "anotherkey": "boo"},)

    assert r.status_code == 200

    r = c.patch_json(obj_link, {"xattrs": {"message": "invalid"}}, expect_errors=True)

    assert r.status_code == 422

    # object output valiidation should bail out with additional xattr
    r = c.get(obj_link)

    assert r.status_code == 200

    # test creation of named object
    r = c.post_json("/named_objects/", {"name": "obj1", "body": "hello"})

    r = c.get("/named_objects/obj1")

    assert r.json["data"]["name"] == "obj1"
    uuid = r.json["data"]["uuid"]
    original_object = r.json["data"]
    r = c.get("/named_objects/+get_uuid?uuid=%s" % uuid)
    object_by_uuid = r.json["data"]
    assert original_object == object_by_uuid

    # duplicate should fail
    r = c.post_json(
        "/named_objects/", {"name": "obj1", "body": "hello"}, expect_errors=True
    )

    assert r.status_code == 422

    # catch issue with ':' in name

    r = c.post_json("/named_objects/", {"name": "object:obj2", "body": "hello"})

    r = c.get("/named_objects/object:obj2")

    assert r.json["data"]["name"] == "object:obj2"

    r = c.get("/named_objects/object:obj2?select=$.[body]")

    assert r.json == ["hello"]

    # catch issue with ' ' in name

    r = c.post_json("/named_objects/", {"name": "object obj2", "body": "hello"})

    r = c.get("/named_objects/object%20obj2")

    assert r.json["data"]["name"] == "object obj2"

    r = c.patch_json("/named_objects/object%20obj2", {"body": "hello1"})

    assert r.status_code == 200

    # blob upload test

    r = c.post_json("/blob_objects", {})

    bloburl = r.json["links"][0]["href"]

    r = c.get(bloburl)

    testimg = os.path.join(os.path.dirname(__file__), "testimg.png")

    r = c.post(bloburl + "/+blobs?field=blobfile", upload_files=[("upload", testimg)])

    assert r.status_code == 200

    r = c.get(bloburl)

    assert r.json["data"]["blobs"]["blobfile"]

    r = c.get(bloburl + "/+blobs?field=blobfile")

    with open(testimg, "rb") as ti:
        assert r.body == ti.read()

    assert r.headers.get("Content-Type") == "image/png"

    r = c.delete(bloburl + "/+blobs?field=blobfile")

    assert r.status_code == 200

    r = c.get(bloburl + "/+blobs?field=blobfile", expect_errors=True)

    assert r.status_code == 404

    # unocide textfile upload test

    r = c.post_json("/blob_objects", {})

    bloburl = r.json["links"][0]["href"]

    r = c.get(bloburl)

    testimg = os.path.join(os.path.dirname(__file__), "testtxt.txt")

    r = c.post(bloburl + "/+blobs?field=blobfile", upload_files=[("upload", testimg)])

    assert r.status_code == 200

    r = c.get(bloburl)

    assert r.json["data"]["blobs"]["blobfile"]

    r = c.get(bloburl + "/+blobs?field=blobfile")

    with open(testimg, "rb") as ti:
        assert r.body == ti.read()

    assert r.headers.get("Content-Type") == "text/plain; charset=UTF-8"

    if os.path.exists(FSBLOB_DIR):
        shutil.rmtree(FSBLOB_DIR)



import typing
from uuid import uuid4

import ldap
import ldap.ldapobject
import webob.exc

from .sqlstorage.sqlstorage import UserSQLStorage

_marker = object()


class LDAP3SQLUserStorage(UserSQLStorage):
    def __init__(
        self,
        request,
        ldap_url: str,
        base_dn: str,
        bind_dn: typing.Optional[str] = None,
        bind_password: typing.Optional[str] = None,
        search_scope=ldap.SCOPE_SUBTREE,
        start_tls: bool = True,
        username_attr: str = "uid",
        email_attr: str = "mail",
        attributes: str = None,
        filterstr: str = r"(&(objectClass=inetOrgPerson)(uid={username}))",
        blobstorage=None,
    ):
        self.ldap_base_dn = base_dn
        self.ldap_url = ldap_url
        self.ldap_start_tls = start_tls
        self.ldap_username_attr = username_attr
        self.ldap_email_attr = email_attr
        self.ldap_search_scope = search_scope
        self.ldap_bind_dn = bind_dn
        self.ldap_bind_password = bind_password
        self.ldap_attr_mapping = [
            v.split("=") for v in (attributes or "").strip().split(",") if v
        ]
        filterstr = filterstr.strip()
        if not filterstr.startswith("(") and not filterstr.endswith(")"):
            filterstr = "(%s)" % filterstr
        self.ldap_filterstr = filterstr
        self.ldap_default_timezone = "UTC"
        super().__init__(request, blobstorage=blobstorage)

    def ldap_connect(self):
        conn = ldap.initialize(self.ldap_url)
        # disable referral chasing
        # https://www.python-ldap.org/en/python-ldap-3.3.0/faq.html
        conn.set_option(ldap.OPT_REFERRALS, 0)
        if self.ldap_start_tls:
            conn.start_tls_s()
        return conn

    def ldap_admin_connect(self):
        client = self.ldap_connect()
        client.bind_s(self.ldap_bind_dn, self.ldap_bind_password)
        return client

    def ldap_get_user(self, username):
        attrs = [self.ldap_username_attr, self.ldap_email_attr] + [
            x[0] for x in self.ldap_attr_mapping
        ]

        ldap_client = self.ldap_admin_connect()
        try:
            result = ldap_client.search_ext_s(
                self.ldap_base_dn,
                self.ldap_search_scope,
                filterstr=self.ldap_filterstr.replace(r"{username}", username),
                attrlist=attrs,
                sizelimit=100,
            )
        except ldap.NO_SUCH_OBJECT:
            return None

        if not result:
            return None

        if result[0][0] is None:
            return None

        userdata = {"dn": result[0][0]}
        for attr in attrs:
            userdata[attr] = result[0][1][attr]
        return userdata

    def get_by_username(self, collection, username):
        user = super().get_by_username(collection, username)
        if user is None:
            ldapuser = self.ldap_get_user(username)
            if ldapuser:
                random_passwd = uuid4().hex
                user = collection.create(
                    {
                        "username": ldapuser[self.ldap_username_attr][0]
                        .decode("utf8")
                        .lower(),
                        "email": ldapuser[self.ldap_email_attr][0]
                        .decode("utf8")
                        .lower(),
                        "source": "ldap",
                        "timezone": self.ldap_default_timezone,
                        "password": random_passwd,
                    },
                )
        return user

    def change_password(self, collection, userid, new_password):
        user = super().get_by_userid(collection, userid)
        if user["source"] == "ldap":
            raise webob.exc.HTTPForbidden()
        return super().change_password(collection, userid, new_password)

    def validate(self, collection, userid, password):
        user = super().get_by_userid(collection, userid)
        if user["source"] == "ldap":
            ldapuser = self.ldap_get_user(user["username"])
            if not ldapuser:
                return False
            dn = ldapuser["dn"]
            ldap_client = self.ldap_connect()
            try:
                ldap_client.bind_s(dn, password)
            except ldap.INVALID_CREDENTIALS:
                return False
            return True
        return super().validate(collection, userid, password)



import re

from .group.path import get_group_collection

UUID_REGEX = re.compile(r"^([a-f0-9]){32}$")


def rellink(context, request, name=None, method="GET", link_name=None):
    link_name = link_name or name
    if name:
        name = "+%s" % name
    if name is None:
        link_name = "self"
    return {"rel": link_name, "type": method, "href": request.link(context, name)}


def has_role(request, role, userid=None, groupname="__default__"):
    if groupname == "__default__" and role == "administrator":
        # FIXME: this is a workaround hack with PAS
        # it should check for which authentication plugin being used
        if userid and UUID_REGEX.match(userid):
            users = request.get_collection("morpfw.pas.user")
            user = users.get_by_userid(userid)
            if user and user["is_administrator"]:
                return True
            return False

    group_col = get_group_collection(request)
    defgroup = group_col.get_by_groupname(groupname)
    if not userid:
        userid = request.identity.userid
    roles = defgroup.get_member_roles(userid)
    users = request.get_collection("morpfw.pas.user")
    user = users.get_by_userid(userid)
    return role in roles



import typing

from inverter import dc2jsl
from jsonschema import ValidationError, validate

from ...interfaces import IXattrProvider

_marker = object()


class XattrProvider(IXattrProvider):

    schema: typing.Type[typing.Any]
    additional_properties = False

    def __init__(self, context):
        self.context = context
        self.request = context.request
        self.app = context.request.app

    def jsonschema(self):
        schema = dc2jsl.convert(
            self.schema,
            ignore_required=True,
            additional_properties=self.additional_properties,
        ).get_schema(ordered=True)
        return {"schema": schema}

    def process_update(self, newdata: dict):
        data = self.as_json()
        data.update(newdata)
        self.schema.validate(self.request, data, context=self.context)
        self.update(data)



from morpfw.crud import permission
from morpfw.crud.errors import UnprocessableError
from morpfw.crud.validator import get_data, validate_schema
from morpfw.util import get_user
from webob.exc import HTTPForbidden, HTTPInternalServerError, HTTPNotFound

from ..app import App
from ..utils import rellink
from ..validator import validate
from .model import GroupCollection, GroupModel, GroupSchema, MemberSchema


@App.json(model=GroupModel, name="members", permission=permission.View)
def list_members(context, request):
    """Return the list of users in the group."""
    members = context.members()
    return {
        "users": [
            {
                "username": m.data["username"],
                "userid": m.userid,
                "roles": context.get_member_roles(m.userid),
                "links": [rellink(m, request)],
            }
            for m in members
        ]
    }


@App.json(
    model=GroupModel, name="grant", request_method="POST", permission=permission.Edit
)
def grant_member(context, request):
    """Grant member roles in the group."""
    mapping = request.json["mapping"]
    for entry in mapping:
        user = entry["user"]
        roles = entry["roles"]
        username = user.get("username", None)
        userid = user.get("userid", None)
        if userid:
            u = context.get_user_by_userid(userid)
        elif username:
            u = context.get_user_by_username(username)
        else:
            u = None
        if u is None:
            raise UnprocessableError("User %s does not exists" % (userid or username))
        for rolename in roles:
            context.grant_member_role(u.userid, rolename)
    return {"status": "success"}


@App.json(
    model=GroupModel, name="revoke", request_method="POST", permission=permission.Edit
)
def revoke_member(context, request):
    """Revoke member roles in the group."""

    mapping = request.json["mapping"]
    for entry in mapping:
        user = entry["user"]
        roles = entry["roles"]
        username = user.get("username", None)
        userid = user.get("userid", None)
        if userid:
            u = context.get_user_by_userid(userid)
        elif username:
            u = context.get_user_by_username(username)
        else:
            u = None
        if u is None:
            raise UnprocessableError("User %s does not exists" % (userid or username))
        for rolename in roles:
            context.revoke_member_role(u.userid, rolename)
    return {"status": "success"}



import time

import morpfw
import pytest

# lets setup a skeleton app

class App(morpfw.BaseApp):
    pass


class Root(object):
    pass


@App.path(model=Root, path='')
def get_root(request):
    return Root()


# lets hook up some scheduled job

# run this code every 5 seconds
@App.periodic(name='myproject.every-5-seconds', seconds=5)
def run_5_secs(request_options):
    print('periodic tick!')


# run this code every 1 minute, using cron style scheduling
@App.cron(name='myproject.minutely', minute='*')
def run_every_1_minute(request_options):
    print('cron tick!')



from ....crud.rulesprovider.base import RulesProvider
from .. import exc
from ..app import App
from ..utils import has_role
from .model import UserCollection, UserModel


class UserRulesProvider(RulesProvider):

    context: UserModel

    def change_password(self, password: str, new_password: str, secure: bool = True):
        context = self.context
        if secure and not has_role(self.request, "administrator"):
            if not context.validate(password, check_state=False):
                raise exc.InvalidPasswordError(context.userid)
        context.storage.change_password(context, context.identity.userid, new_password)

    def validate(self, password: str, check_state=True) -> bool:
        context = self.context
        if check_state and context.data["state"] != "active":
            return False
        return context.storage.validate(context, context.userid, password)


@App.rulesprovider(model=UserModel)
def get_user_rulesprovider(context):
    return UserRulesProvider(context)



import rulez
from morpfw.crud import Collection, Model, Schema
from morpfw.crud import errors as cruderrors

from .. import exc
from ..app import App
from ..exc import GroupExistsError
from ..model import NAME_PATTERN
from .schema import GroupSchema, MemberSchema

DEFAULT_VALID_ROLES = ["member", "administrator"]


class GroupCollection(Collection):
    schema = GroupSchema
    exist_exc = GroupExistsError

    def get_by_groupname(self, groupname):
        return self.storage.get_by_groupname(self, groupname)


class GroupModel(Model):
    schema = GroupSchema

    def get_user_by_userid(self, userid):
        return self.storage.get_user_by_userid(self, userid)

    def get_user_by_username(self, username):
        return self.storage.get_user_by_username(self, username)

    def members(self):
        members = self.storage.get_members(self, self.identifier)
        active_members = [
            member for member in members if member.data.get("state") == "active"
        ]
        return active_members

    def children(self):
        col = GroupCollection(self.request, self.storage)
        children = col.search(rulez.field["parent"] == self.identifier)
        return children

    def add_members(self, userids):
        self.storage.add_group_members(self, self.identifier, userids)
        for userid in userids:
            self.storage.grant_group_user_role(self, self.identifier, userid, "member")

    def remove_members(self, userids):
        self.storage.remove_group_members(self, self.identifier, userids)

    def get_member_roles(self, userid):
        return self.storage.get_group_user_roles(self, self.identifier, userid)

    def grant_member_role(self, userid, rolename):
        if userid not in [m.userid for m in self.members()]:
            self.add_members([userid])
        valid_roles = self.request.app.get_config(
            "morpfw.valid_roles", DEFAULT_VALID_ROLES
        )
        if rolename not in valid_roles:
            raise exc.InvalidRoleError(rolename)
        self.storage.grant_group_user_role(self, self.identifier, userid, rolename)

    def revoke_member_role(self, userid, rolename):
        if rolename == "member":
            self.remove_members([userid])
            return
        self.storage.revoke_group_user_role(self, self.identifier, userid, rolename)
        if not self.get_member_roles(userid) and self.identifier != "__default__":
            self.remove_members([userid])

    def _links(self):
        links = []
        for m in self.members():
            links.append({"rel": "member", "href": self.request.link(m)})
        return links



from urllib.parse import urlencode
import json
import requests


qs = urlencode({
    'group': ("count:count(uuid), year:year(created), month:month(created),"
              "day:day(created), sum:sum(value), avg:avg(value)")
})
requests.get('/pages/+aggregate?%s' % qs)



import re
import urllib
from functools import partial

import reg
from jsonschema import Draft4Validator
from morepath.publish import resolve_model

from .errors import FieldValidationError, FormValidationError, ValidationError


@reg.dispatch(reg.match_instance("model"), reg.match_instance("request"))
def get_data(model, request):
    raise NotImplementedError


def regex_validator(pattern, name):
    p = re.compile(pattern)

    def _regex_validator(request, schema, field, value, mode=None):
        if not p.match(value):
            return "%s does not match %s pattern" % (value, name)

    return _regex_validator


def load(validator, schema, request):
    newreq = request.app.request_class(
        request.environ.copy(),
        request.app.root,
        path_info=urllib.parse.unquote(request.path),
    )
    context = resolve_model(newreq)
    context.request = request
    if schema is None:
        dc = context.schema
    else:
        dc = schema

    data = get_data(context, request)
    return dc.validate(request, data, context=context)


def validate_schema(validator=Draft4Validator, schema=None):
    return partial(load, validator, schema)



import morepath

from ..app import App
from .model import CurrentUserModel, UserCollection, UserModel, UserSchema


def get_user(request: morepath.Request, identifier) -> UserModel:
    collection = get_user_collection(request)
    return collection.get(identifier)


def get_user_collection(request: morepath.Request) -> UserCollection:
    return UserCollection(request, storage=request.app.get_storage(UserModel, request))


def get_current_user(request: morepath.Request) -> UserModel:
    userid = request.identity.userid
    collection = get_user_collection(request)
    return collection.get_by_userid(userid)


def refresh_nonce_handler(request, userid):
    collection = get_user_collection(request)
    return collection.get_by_userid(userid)["nonce"]



import morepath
import reg

from .model import Collection, Model


class WorkflowCollection(Collection):
    pass


class WorkflowModel(Model):
    @classmethod
    def transition(klass, model):
        def apply_transition(func):
            klass._apply_transition.subscribe(obj=model, request=morepath.Request)(func)

        return apply_transition

    @reg.dispatch_method(
        reg.match_instance("obj"), reg.match_instance("request"),
    )
    def _apply_transition(self, obj, request):
        pass

    def process(self, obj):
        self._apply_transition.publish(self, obj, self.request)



import morepath
from .app import App
from . import subscribers
import argparse
import yaml
import sqlalchemy
import os
from .model import Collection, Model
from .rulesprovider.base import RulesProvider
from .schema import Schema
from .statemachine.base import StateMachine
from .aggregateprovider.base import AggregateProvider
from .searchprovider.base import SearchProvider
from .util import resolve_model
from .app import App
from .storage.sqlstorage import SQLStorage



import typing
from dataclasses import dataclass, field

import morpfw
import morpfw.sql
import sqlalchemy as sa
from morpfw.authz.pas import DefaultAuthzPolicy
from morpfw.crud import permission as crudperm
from morpfw.permission import All


class AppRoot(object):
    def __init__(self, request):
        self.request = request


class App(DefaultAuthzPolicy, morpfw.SQLApp):
    pass


@App.path(model=AppRoot, path="/")
def get_approot(request):
    return AppRoot(request)


@App.permission_rule(model=AppRoot, permission=All)
def allow_all(identity, context, permission):
    """ Default permission rule, allow all """
    return True


@App.json(model=AppRoot)
def index(context, request):
    return {"message": "Hello World"}


@dataclass
class PageSchema(morpfw.Schema):

    body: typing.Optional[str] = field(default=None, metadata={"title": "Body"})
    value: typing.Optional[int] = field(default=0, metadata={"title": "Value"})


class PageCollection(morpfw.Collection):
    schema = PageSchema


class PageModel(morpfw.Model):
    schema = PageSchema

    blob_fields = ['attachment']

# SQLALchemy model
class Page(morpfw.sql.Base):

    __tablename__ = "test_page"

    body = sa.Column(sa.Text())
    value = sa.Collection(sa.Integer())


class PageStorage(morpfw.SQLStorage):
    model = PageModel
    orm_model = Page


@App.storage(model=PageModel)
def get_storage(model, request, blobstorage):
    return PageStorage(request, blobstorage=blobstorage)


@App.path(model=PageCollection, path="/pages")
def get_collection(request):
    storage = request.app.get_storage(PageModel, request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path="/pages/{identifier}")
def get_model(request, identifier):
    col = get_collection(request)
    return col.get(identifier)


class PageStateMachine(morpfw.StateMachine):

    states = ["new", "pending", "approved"]
    transitions = [
        {"trigger": "approve", "source": ["new", "pending"], "dest": "approved"},
        {"trigger": "submit", "source": "new", "dest": "pending"},
    ]


@App.statemachine(model=PageModel)
def get_pagemodel_statemachine(context):
    return PageStateMachine(context)


@App.permission_rule(model=PageCollection, permission=All)
def allow_collection_all(identity, context, permission):
    """ Default permission rule, allow all """
    return True


@App.permission_rule(model=PageModel, permission=All)
def allow_model_all(identity, context, permission):
    """ Default permission rule, allow all """
    return True


@App.typeinfo(name="test.page", schema=PageSchema)
def get_typeinfo(request):
    return {
        "title": "Test Page",
        "description": "",
        "schema": PageSchema,
        "collection": PageCollection,
        "collection_factory": get_collection,
        "model": PageModel,
        "model_factory": get_model,
    }



import uuid

import morepath
import rulez
from morpfw.crud import permission

from ..app import App
from ..path import get_apikey, get_apikey_collection, get_user, get_user_collection
from .model import APIKeyCollection, APIKeyModel


@App.json(
    model=APIKeyCollection,
    name="generate",
    request_method="POST",
    permission=permission.Create,
)
def generate_apikey(context, request):
    usercol = get_user_collection(request)
    user = usercol.get_by_userid(request.identity.userid)
    data = {}
    if not user.validate(request.json["password"]):

        @request.after
        def adjust_response(response):
            response.status = 422

        return {"status": "error", "message": "Invalid password"}

    data["name"] = request.json["name"]
    obj = context.create(data)
    secret = obj.generate_secret()
    result = obj.json()
    result["data"]["api_secret"] = secret
    return result



import typing
import morpfw
from .app import App
from .model import PageCollection


class PagesAggregateProvider(morpfw.AggregateProvider):

    def aggregate(self, query=None, group=None, order_by=None):
        """search for resources, aggregate and return aggregate result"""
        result = []
        # do something here
        return result


@App.aggregateprovider(model=PageCollection)
def get_aggregateprovider(context):
    return PagesAggregateProvider(context)
